import re
import subprocess
import warnings
from pathlib import Path

# This import is needed for alembic to recognize the geopackage dialect
import geoalchemy2.alembic_helpers  # noqa: F401
from alembic import command as alembic_command
from alembic.config import Config
from alembic.environment import EnvironmentContext
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from sqlalchemy import Column, Integer, MetaData, Table, text
from sqlalchemy.exc import IntegrityError

from ..domain import constants, models
from ..infrastructure.spatial_index import ensure_spatial_indexes
from ..infrastructure.spatialite_versions import copy_models, get_spatialite_version
from .errors import MigrationMissingError, UpgradeFailedError
from .upgrade_utils import setup_logging

__all__ = ["ModelSchema"]


def get_alembic_config(engine=None, unsafe=False):
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", "threedi_schema:migrations")
    alembic_cfg.set_main_option("version_table", constants.VERSION_TABLE_NAME)
    if engine is not None:
        alembic_cfg.attributes["engine"] = engine
    alembic_cfg.attributes["unsafe"] = unsafe
    return alembic_cfg


def get_schema_version():
    """Returns the version of the schema in this library"""
    config = get_alembic_config()
    script = ScriptDirectory.from_config(config)
    with EnvironmentContext(config=config, script=script) as env:
        return int(env.get_head_revision())


def _upgrade_database(db, revision="head", unsafe=True, progress_func=None):
    """Upgrade ThreediDatabase instance"""
    engine = db.engine
    config = get_alembic_config(engine, unsafe=unsafe)
    if progress_func is not None:
        setup_logging(db.schema, revision, config, progress_func)
    alembic_command.upgrade(config, revision)


class ModelSchema:
    def __init__(self, threedi_db, declared_models=models.DECLARED_MODELS):
        self.db = threedi_db
        self.declared_models = declared_models

    def _get_version_old(self):
        """The version of the database using the old 'south' versioning."""
        south_migrationhistory = Table(
            "south_migrationhistory", MetaData(), Column("id", Integer)
        )
        engine = self.db.engine
        if not self.db.has_table("south_migrationhistory"):
            return
        with engine.connect() as connection:
            query = south_migrationhistory.select().order_by(
                south_migrationhistory.columns["id"].desc()
            )
            versions = list(connection.execute(query.limit(1)))
            if len(versions) == 1:
                return versions[0][0]
            else:
                return None

    def get_version(self):
        """Returns the id (integer) of the latest migration"""
        with self.db.engine.connect() as connection:
            context = MigrationContext.configure(
                connection, opts={"version_table": constants.VERSION_TABLE_NAME}
            )
            version = context.get_current_revision()
        if version is not None:
            return int(version)
        else:
            return self._get_version_old()

    def upgrade(
        self,
        revision="head",
        backup=True,
        upgrade_spatialite_version=False,
        convert_to_geopackage=False,
        progress_func=None,
    ):
        """Upgrade the database to the latest version.

        This requires either a completely empty database or a database with its
        current schema version at least 174 (the latest migration of the old
        model databank).

        The upgrade is done using database transactions. However, for SQLite,
        database transactions are only partially supported. To ensure that the
        database file does not become corrupt, enable the "backup" parameter.
        If the database is temporary already (or if it is PostGIS), disable
        it.

        Specify 'upgrade_spatialite_version=True' to also upgrade the
        spatialite file version after the upgrade.

        Specify 'convert_to_geopackage=True' to also convert from spatialite
        to geopackage file version after the upgrade.

        Specify a 'progress_func' to handle progress updates. `progress_func` should
        expect a single argument representing the fraction of progress
        """
        try:
            rev_nr = get_schema_version() if revision == "head" else int(revision)
        except ValueError:
            raise ValueError(
                f"Incorrect version format: {revision}. Expected 'head' or a numeric value."
            )
        if convert_to_geopackage and rev_nr < 300:
            raise UpgradeFailedError(
                f"Cannot convert to geopackage for {revision=} because geopackage support is "
                "enabled from revision 300",
            )
        v = self.get_version()
        if v is not None and v < constants.LATEST_SOUTH_MIGRATION_ID:
            raise MigrationMissingError(
                f"This tool cannot update versions below "
                f"{constants.LATEST_SOUTH_MIGRATION_ID}. Please consult the "
                f"3Di documentation on how to update legacy databases."
            )
        if backup:
            with self.db.file_transaction() as work_db:
                _upgrade_database(
                    work_db, revision=revision, unsafe=True, progress_func=progress_func
                )
        else:
            _upgrade_database(
                self.db, revision=revision, unsafe=False, progress_func=progress_func
            )
        if upgrade_spatialite_version:
            self.upgrade_spatialite_version()
        elif convert_to_geopackage:
            self.convert_to_geopackage()

    def validate_schema(self):
        """Very basic validation of 3Di schema.

        Check that the database has the latest migration applied. If the
        latest migrations is applied, we assume the database also contains all
        tables and columns defined in threedi_model.models.py.

        :return: True if the threedi_db schema is valid, raises an error otherwise.
        :raise MigrationMissingError, MigrationTooHighError
        """
        version = self.get_version()
        schema_version = get_schema_version()
        if version is None or version < schema_version:
            raise MigrationMissingError(
                f"This tool requires at least schema version "
                f"{schema_version}. Current version: {version}."
            )

        if version > schema_version:
            warnings.warn(
                f"The database version is higher than the threedi-schema version "
                f"({version} > {schema_version}). This may lead to unexpected "
                f"results. "
            )
        return True

    def set_spatial_indexes(self):
        """(Re)create spatial indexes in the spatialite according to the latest definitions."""
        version = self.get_version()
        schema_version = get_schema_version()
        if version != schema_version:
            raise MigrationMissingError(
                f"Setting views requires schema version "
                f"{schema_version}. Current version: {version}."
            )

        ensure_spatial_indexes(self.db, models.DECLARED_MODELS)

    def upgrade_spatialite_version(self):
        """Upgrade the version of the spatialite file to the version of the
        current spatialite library.

        Does nothing if the current file version > 3 or if the current library
        version is not 4 or 5.

        Raises UpgradeFailedError if there are any SQL constraints violated.
        """
        lib_version, file_version = get_spatialite_version(self.db)
        if file_version == 3 and lib_version in (4, 5):
            self.validate_schema()

            with self.db.file_transaction(start_empty=True) as work_db:
                _upgrade_database(work_db, revision="head", unsafe=True)
                try:
                    copy_models(self.db, work_db, self.declared_models)
                except IntegrityError as e:
                    raise UpgradeFailedError(e.orig.args[0])

    def convert_to_geopackage(self):
        """
        Convert spatialite to geopackage using gdal's ogr2ogr.

        Does nothing if the current database is already a geopackage.

        Raises UpgradeFailedError if the conversion of spatialite to geopackage with ogr2ogr fails.
        """
        if self.db.get_engine().dialect.name == "geopackage":
            return
        # Check if ogr2ogr
        result = subprocess.run(
            "ogr2ogr --version",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        ## ogr2ogr is installed; make sure the version is high enough and return if not
        if result.returncode == 0:
            # get version
            version = re.findall(r"\b(\d+\.\d+\.\d+)\b", result.stdout)[0]
            # trim patch version and convert to float
            float_version = float(version[0 : version.rfind(".")])
            if float_version < 3.4:
                warnings.warn(
                    f"ogr2ogr 3.4 (part of GDAL) or newer is needed to convert spatialite to geopackage "
                    f"but ogr2ogr {version} was found. {self.db.path} will not be converted"
                    f"to geopackage."
                )
                return
        # ogr2ogr is not (properly) installed; return
        elif result.returncode != 0:
            warnings.warn(
                f"ogr2ogr (part of GDAL) is needed to convert spatialite to geopackage but no working"
                f"working installation was found:\n{result.stderr}"
            )
            return
        # Ensure database is upgraded and views are recreated
        self.upgrade()
        self.validate_schema()
        # Make necessary modifications for conversion on temporary database
        with self.db.file_transaction(start_empty=False, copy_results=False) as work_db:
            # remove spatialite specific tables that break conversion
            with work_db.get_session() as session:
                session.execute(text("DROP TABLE IF EXISTS spatialite_history;"))
                session.execute(text("DROP TABLE IF EXISTS views_geometry_columns;"))
            cmd = [
                "ogr2ogr",
                "-skipfailures",
                "-f",
                "gpkg",
                str(Path(self.db.path).with_suffix(".gpkg")),
                str(work_db.path),
                "-oo",
                "LIST_ALL_TABLES=YES",
            ]
            try:
                p = subprocess.Popen(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=-1
                )
            except Exception as e:
                raise UpgradeFailedError(f"ogr2ogr failed conversion:\n{e}")
            _, out = p.communicate()
        # Error handling
        # convert bytes to utf and split lines
        out_list = out.decode("utf-8").split("\n")
        # collect only errors and remove 'ERROR #:'
        errors = [
            [idx, ": ".join(item.split(": ")[1:])]
            for idx, item in enumerate(out_list)
            if item.lower().startswith("error")
        ]
        # While creating the geopackage with ogr2ogr an error occurs
        # because ogr2ogr tries to create a table `sqlite_sequence`, which
        # is reserved for internal use. The resulting database seems fine,
        # so this specific error is ignored
        # convert error output to list
        expected_error = 'sqlite3_exec(CREATE TABLE "sqlite_sequence" ( "rowid" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, "name" TEXT, "seq" TEXT)) failed: object name reserved for internal use: sqlite_sequence'
        unexpected_error_indices = [
            idx for idx, error in errors if error.lower() != expected_error.lower()
        ]
        if len(unexpected_error_indices) > 0:
            error_str = "\n".join(
                [out_list[idx].decode("utf-8") for idx in unexpected_error_indices]
            )
            raise UpgradeFailedError(f"ogr2ogr didn't finish as expected:\n{error_str}")
        # Correct path of current database
        self.db.path = Path(self.db.path).with_suffix(".gpkg")
        # Reset engine so new path is used on the next call of get_engine()
        self.db._engine = None
        # Recreate views_geometry_columns so set_views works as expected
        with self.db.get_session() as session:
            session.execute(
                text(
                    "CREATE TABLE views_geometry_columns(view_name TEXT, view_geometry TEXT, view_rowid TEXT, f_table_name VARCHAR(256), f_geometry_column VARCHAR(256))"
                )
            )



from unittest import mock

import pytest
from sqlalchemy import Column, Integer, MetaData, String, Table, text

from threedi_schema import ModelSchema
from threedi_schema.application import errors
from threedi_schema.application.schema import get_schema_version
from threedi_schema.domain import constants
from threedi_schema.infrastructure.spatialite_versions import get_spatialite_version


@pytest.fixture
def south_migration_table(in_memory_sqlite):
    south_migrationhistory = Table(
        "south_migrationhistory", MetaData(), Column("id", Integer)
    )
    engine = in_memory_sqlite.engine
    south_migrationhistory.create(engine)
    return south_migrationhistory


@pytest.fixture
def alembic_version_table(in_memory_sqlite):
    alembic_version = Table(
        constants.VERSION_TABLE_NAME,
        MetaData(),
        Column("version_num", String(32), nullable=False),
    )
    engine = in_memory_sqlite.engine
    alembic_version.create(engine)
    return alembic_version


def test_get_schema_version():
    """The current version in the library. We start counting at 200."""
    # this will catch future mistakes of setting non-integer revisions
    assert get_schema_version() >= 200


def test_get_version_no_tables(in_memory_sqlite):
    """Get the version of a sqlite with no version tables"""
    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id is None


def test_get_version_empty_south(in_memory_sqlite, south_migration_table):
    """Get the version of a sqlite with an empty South version table"""
    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id is None


def test_get_version_south(in_memory_sqlite, south_migration_table):
    """Get the version of a sqlite with a South version table"""
    with in_memory_sqlite.engine.connect() as connection:
        with connection.begin():
            for v in (42, 43):
                connection.execute(south_migration_table.insert().values(id=v))

    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id == 43


def test_get_version_empty_alembic(in_memory_sqlite, alembic_version_table):
    """Get the version of a sqlite with an empty alembic version table"""
    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id is None


def test_get_version_alembic(in_memory_sqlite, alembic_version_table):
    """Get the version of a sqlite with an alembic version table"""
    with in_memory_sqlite.engine.connect() as connection:
        with connection.begin():
            connection.execute(
                alembic_version_table.insert().values(version_num="0201")
            )

    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id == 201


def test_validate_schema(sqlite_latest):
    """Validate a correct schema version"""
    schema = sqlite_latest.schema
    with mock.patch.object(schema, "get_version", return_value=get_schema_version()):
        assert schema.validate_schema()


@pytest.mark.parametrize("version", [-1, 205, None])
def test_validate_schema_missing_migration(sqlite_latest, version):
    """Validate a too low schema version"""
    schema = ModelSchema(sqlite_latest)
    with mock.patch.object(schema, "get_version", return_value=version):
        with pytest.raises(errors.MigrationMissingError):
            schema.validate_schema()


@pytest.mark.parametrize("version", [9999])
def test_validate_schema_too_high_migration(sqlite_latest, version):
    """Validate a too high schema version"""
    schema = ModelSchema(sqlite_latest)
    with mock.patch.object(schema, "get_version", return_value=version):
        with pytest.warns(UserWarning):
            schema.validate_schema()


def test_full_upgrade_empty(in_memory_sqlite):
    """Upgrade an empty database to the latest version"""
    schema = ModelSchema(in_memory_sqlite)
    schema.upgrade(backup=False, upgrade_spatialite_version=False)
    assert schema.get_version() == get_schema_version()
    assert in_memory_sqlite.has_table("connection_node")


def test_full_upgrade_with_preexisting_version(south_latest_sqlite):
    """Upgrade an empty database to the latest version"""
    schema = ModelSchema(south_latest_sqlite)
    schema.upgrade(backup=False, upgrade_spatialite_version=False)
    assert schema.get_version() == get_schema_version()
    assert south_latest_sqlite.has_table("connection_node")
    # https://github.com/nens/threedi-schema/issues/10:
    assert not south_latest_sqlite.has_table("v2_levee")


def test_full_upgrade_oldest(oldest_sqlite):
    """Upgrade a legacy database to the latest version"""
    schema = ModelSchema(oldest_sqlite)
    schema.upgrade(backup=False, upgrade_spatialite_version=False)
    assert schema.get_version() == get_schema_version()
    assert oldest_sqlite.has_table("connection_node")
    # https://github.com/nens/threedi-schema/issues/10:
    assert not oldest_sqlite.has_table("v2_levee")


def test_upgrade_south_not_latest_errors(in_memory_sqlite):
    """Upgrading a database that is not at the latest south migration will error"""
    schema = ModelSchema(in_memory_sqlite)
    with mock.patch.object(
        schema, "get_version", return_value=constants.LATEST_SOUTH_MIGRATION_ID - 1
    ):
        with pytest.raises(errors.MigrationMissingError):
            schema.upgrade(backup=False, upgrade_spatialite_version=False)


def test_upgrade_with_backup(south_latest_sqlite):
    """Upgrading with backup=True will proceed on a copy of the database"""
    schema = ModelSchema(south_latest_sqlite)
    with mock.patch(
        "threedi_schema.application.schema._upgrade_database", side_effect=RuntimeError
    ) as upgrade, mock.patch.object(schema, "get_version", return_value=199):
        with pytest.raises(RuntimeError):
            schema.upgrade(backup=True, upgrade_spatialite_version=False)

    (db,), kwargs = upgrade.call_args
    assert db is not south_latest_sqlite


def test_upgrade_without_backup(south_latest_sqlite):
    """Upgrading with backup=True will proceed on the database itself"""
    schema = ModelSchema(south_latest_sqlite)
    with mock.patch(
        "threedi_schema.application.schema._upgrade_database", side_effect=RuntimeError
    ) as upgrade, mock.patch.object(schema, "get_version", return_value=199):
        with pytest.raises(RuntimeError):
            schema.upgrade(backup=False, upgrade_spatialite_version=False)

    (db,), kwargs = upgrade.call_args
    assert db is south_latest_sqlite


def test_convert_to_geopackage_raise(oldest_sqlite):
    if get_schema_version() >= 300:
        pytest.skip("Warning not expected beyond schema 300")
    schema = ModelSchema(oldest_sqlite)
    with pytest.raises(errors.UpgradeFailedError):
        schema.upgrade(
            backup=False, upgrade_spatialite_version=False, convert_to_geopackage=True
        )


def test_upgrade_revision_exception(oldest_sqlite):
    schema = ModelSchema(oldest_sqlite)
    with pytest.raises(ValueError):
        schema.upgrade(revision="foo")


def test_upgrade_spatialite_3(oldest_sqlite):
    lib_version, file_version_before = get_spatialite_version(oldest_sqlite)
    if lib_version == file_version_before:
        pytest.skip("Nothing to test: spatialite library version equals file version")

    schema = ModelSchema(oldest_sqlite)
    schema.upgrade(backup=False, upgrade_spatialite_version=True)

    _, file_version_after = get_spatialite_version(oldest_sqlite)
    assert file_version_after == 4

    # the spatial indexes are there
    with oldest_sqlite.engine.connect() as connection:
        check_result = connection.execute(
            text("SELECT CheckSpatialIndex('connection_node', 'geom')")
        ).scalar()
    assert check_result == 1


def test_set_spatial_indexes(in_memory_sqlite):
    engine = in_memory_sqlite.engine

    schema = ModelSchema(in_memory_sqlite)
    schema.upgrade(backup=False)

    with engine.connect() as connection:
        with connection.begin():
            connection.execute(
                text("SELECT DisableSpatialIndex('connection_node', 'geom')")
            ).scalar()
            connection.execute(text("DROP TABLE idx_v2_connection_nodes_the_geom"))

    schema.set_spatial_indexes()

    with engine.connect() as connection:
        check_result = connection.execute(
            text("SELECT CheckSpatialIndex('connection_node', 'geom')")
        ).scalar()

    assert check_result == 1



import csv
import shutil
import sqlite3
from pathlib import Path

import pytest
from geoalchemy2 import Geometry
from sqlalchemy import inspect

from threedi_schema import ModelSchema, ThreediDatabase

# All these tests are marked with the marker `migrations`.
# Individual tests and test classes are additionally marked with a marker relating
# to the specific migration that is being tested, e.g. `migration_300`.
# To run tests for all migrations: pytest -m migrations
# To run tests for a specific migration: pytest -m migration_xyz
# To exclude migration tests: pytest -m "no migrations"

pytestmark = pytest.mark.migrations

data_dir = Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def sqlite_path():
    return data_dir.joinpath("v2_bergermeer_221.sqlite")


@pytest.fixture(scope="session")
def schema_upgraded(tmp_path_factory, sqlite_path):
    tmp_sqlite = tmp_path_factory.mktemp("custom_dir").joinpath(sqlite_path.name)
    shutil.copy(sqlite_path, tmp_sqlite)
    schema = ModelSchema(ThreediDatabase(tmp_sqlite))
    schema.upgrade(backup=False)
    return schema


@pytest.fixture(scope="session")
def schema_ref(tmp_path_factory, sqlite_path):
    # Last pre-upgrade version
    tmp_sqlite = tmp_path_factory.mktemp("custom_dir").joinpath(sqlite_path.name)
    shutil.copy(sqlite_path, tmp_sqlite)
    return ModelSchema(ThreediDatabase(tmp_sqlite))


def get_sql_tables(cursor):
    return [item[0] for item in cursor.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()]


def get_cursor_for_schema(schema):
    return sqlite3.connect(schema.db.path).cursor()


def get_columns_from_schema(schema, table_name):
    inspector = inspect(schema.db.get_engine())
    columns = inspector.get_columns(table_name)
    return {column['name']: (str(column['type']).lower(), column['nullable']) for column in columns
            if not 'geom' in column['name']}


def get_columns_from_sqlite(cursor, table_name):
    cursor.execute(f"PRAGMA table_info({table_name})")
    col_map = {}
    for c in cursor.fetchall():
        if 'geom' in c[1]:
            continue
        type_str = c[2].lower() if c[2] != 'bool' else 'boolean'
        col_map[c[1]] = (type_str, not c[3])
    return col_map


def get_values_from_sqlite(cursor, table_name, column_name):
    cursor.execute(f"SELECT {column_name} FROM {table_name}")
    return cursor.fetchall()


@pytest.mark.parametrize("sqlite_file",
                         ["v2_bergermeer_221.sqlite",
                          "staging-test-0d1d2d-simple-infiltration.sqlite",
                          "staging-test-0d1d2d-simple-infiltration_surface.sqlite"])
def test_upgrade_success(sqlite_file, tmp_path_factory):
    tmp_sqlite = tmp_path_factory.mktemp("custom_dir").joinpath(sqlite_file)
    shutil.copy(data_dir.joinpath(sqlite_file), tmp_sqlite)
    schema = ModelSchema(ThreediDatabase(tmp_sqlite))
    # Test if running upgrade doesn't run into any exceptions
    try:
        schema.upgrade(backup=False)
    except Exception:
        pytest.fail(f"Failed to upgrade {sqlite_file}")


class TestMigration228:
    pytestmark = pytest.mark.migration_228
    removed_tables = set(["v2_channel",
                      "v2_windshielding",
                      "v2_cross_section_location",
                      "v2_pipe",
                      "v2_culvert",
                      "v2_weir",
                      "v2_orifice",
                      "v2_pumpstation",
                      "v2_cross_section_definition",
                      "v2_floodfill",
                      "v2_connection_nodes"])
    added_tables = set(["channel",
                    "windshielding_1d",
                    "cross_section_location",
                    "pipe",
                    "culvert",
                    "weir",
                    "orifice",
                    "pump",
                    "connection_node",
                    "material",
                    "pump_map"])

    def test_tables(self, schema_ref, schema_upgraded):
        # Test whether the added tables are present
        # and whether the removed tables are not present*
        tables_new = set(get_sql_tables(get_cursor_for_schema(schema_upgraded)))
        assert self.added_tables.issubset(tables_new)
        assert self.removed_tables.isdisjoint(tables_new)


    def test_columns_added_tables(self, schema_upgraded):
        # Note that only the added tables are touched.
        # So this check covers both added and removed columns.
        cursor = get_cursor_for_schema(schema_upgraded)
        for table in self.added_tables:
            cols_sqlite = get_columns_from_sqlite(cursor, table)
            cols_schema = get_columns_from_schema(schema_upgraded, table)
            assert cols_sqlite == cols_schema


class TestMigration226:
    pytestmark = pytest.mark.migration_226
    removed_tables = set(['v2_dem_average_area',
                          'v2_exchange_line',
                          'v2_grid_refinement',
                          'v2_grid_refinement_area',
                          'v2_obstacle',
                          'v2_potential_breach'])
    added_tables = set(['dem_average_area',
                          'exchange_line',
                          'grid_refinement_line',
                          'grid_refinement_area',
                          'obstacle',
                          'potential_breach'])

    def test_tables(self, schema_ref, schema_upgraded):
        # Test whether the added tables are present
        # and whether the removed tables are not present*
        tables_new = set(get_sql_tables(get_cursor_for_schema(schema_upgraded)))
        assert self.added_tables.issubset(tables_new)
        assert self.removed_tables.isdisjoint(tables_new)


    def test_columns_added_tables(self, schema_upgraded):
        # Note that only the added tables are touched.
        # So this check covers both added and removed columns.
        cursor = get_cursor_for_schema(schema_upgraded)
        for table in self.added_tables:
            cols_sqlite = get_columns_from_sqlite(cursor, table)
            cols_schema = get_columns_from_schema(schema_upgraded, table)
            assert cols_sqlite == cols_schema

class TestMigration225:
    pytestmark = pytest.mark.migration_225
    removed_tables = set(['v2_1d_lateral', 'v2_2d_lateral', 'v2_1d_boundary_conditions',
                          'v2_2d_boundary_conditions'])
    added_tables = set(['lateral_1d', 'lateral_2d', 'boundary_condition_1d', 'boundary_condition_2d'])

    def test_tables(self, schema_ref, schema_upgraded):
        # Test whether the added tables are present
        # and whether the removed tables are not present*
        tables_new = set(get_sql_tables(get_cursor_for_schema(schema_upgraded)))
        assert self.added_tables.issubset(tables_new)
        assert self.removed_tables.isdisjoint(tables_new)

    def test_columns_added_tables(self, schema_upgraded):
        # Note that only the added tables are touched.
        # So this check covers both added and removed columns.
        cursor = get_cursor_for_schema(schema_upgraded)
        for table in self.added_tables:
            cols_sqlite = get_columns_from_sqlite(cursor, table)
            cols_schema = get_columns_from_schema(schema_upgraded, table)
            assert cols_sqlite == cols_schema


class TestMigration224:
    pytestmark = pytest.mark.migration_224
    removed_tables = set(['v2_control', 'v2_control_delta', 'v2_control_group',
                          'v2_control_measure_group', 'v2_control_measure_map',
                          'v2_control_pid', 'v2_control_timed',
                          'v2_control_memory', 'v2_control_table'])
    added_tables = set(['memory_control', 'table_control', 'measure_map', 'measure_location'])

    def test_tables(self, schema_ref, schema_upgraded):
        # Test whether the added tables are present
        # and whether the removed tables are not present*
        tables_new = set(get_sql_tables(get_cursor_for_schema(schema_upgraded)))
        assert self.added_tables.issubset(tables_new)
        assert self.removed_tables.isdisjoint(tables_new)

    def test_columns_added_tables(self, schema_upgraded):
        # Note that only the added tables are touched.
        # So this check covers both added and removed columns.
        cursor = get_cursor_for_schema(schema_upgraded)
        for table in self.added_tables:
            cols_sqlite = get_columns_from_sqlite(cursor, table)
            cols_schema = get_columns_from_schema(schema_upgraded, table)
            assert cols_sqlite == cols_schema


class TestMigration223:
    pytestmark = pytest.mark.migration_223
    removed_tables = set(['v2_surface', 'v2_surface_parameters', 'v2_surface_map',
                          'v2_impervious_surface', 'v2_impervious_surface_map'])
    added_tables = set(['surface', 'surface_map', 'surface_parameters', 'tags',
                        'dry_weather_flow', 'dry_weather_flow_map', 'dry_weather_flow_distribution'])

    def test_tables(self, schema_ref, schema_upgraded):
        # Test whether the added tables are present
        # and whether the removed tables are not present*
        tables_new = set(get_sql_tables(get_cursor_for_schema(schema_upgraded)))
        assert self.added_tables.issubset(tables_new)
        assert self.removed_tables.isdisjoint(tables_new)

    def test_columns_added_tables(self, schema_upgraded):
        # Note that only the added tables are touched.
        # So this check covers both added and removed columns.
        cursor = get_cursor_for_schema(schema_upgraded)
        for table in self.added_tables:
            cols_sqlite = get_columns_from_sqlite(cursor, table)
            cols_schema = get_columns_from_schema(schema_upgraded, table)
            assert cols_sqlite == cols_schema


class TestMigration222:
    pytestmark = pytest.mark.migration_222

    with open(data_dir.joinpath('migration_222.csv'), 'r') as file:
        # src_table, src_column, dst_table, dst_column
        migration_map = [[row[0], row[1], row[2], row[3]] for row in csv.reader(file)]
    removed_tables = set([row[0] for row in migration_map])
    added_tables = set([row[2] for row in migration_map])
    bool_settings_id = [
        ("use_groundwater_storage", "groundwater_settings_id", "groundwater"),
        ("use_interflow", "interflow_settings_id", "interflow"),
        ("use_structure_control", "control_group_id", "v2_control_group"),
        ("use_simple_infiltration", "simple_infiltration_settings_id", "simple_infiltration"),
        ("use_vegetation_drag_2d", "vegetation_drag_settings_id", "vegetation_drag_2d")
    ]
    bool_settings_exist = [
        ("use_interception", "interception")
    ]
    single_row_tables = ["model_settings",
                         "simulation_template_settings",
                         "time_step_settings",
                         "numerical_settings",
                         "physical_settings",
                         "initial_conditions",
                         "interception"]

    def test_tables(self, schema_ref, schema_upgraded):
        # Test whether the added tables are present
        # and whether the removed tables are not present*
        tables_new = set(get_sql_tables(get_cursor_for_schema(schema_upgraded)))
        assert self.added_tables.issubset(tables_new)
        assert self.removed_tables.isdisjoint(tables_new)

    def test_columns_added_tables(self, schema_upgraded):
        # Note that only the added tables are touched.
        # So this check covers both added and removed columns.
        cursor = get_cursor_for_schema(schema_upgraded)
        for table in self.added_tables:
            cols_sqlite = get_columns_from_sqlite(cursor, table)
            cols_schema = get_columns_from_schema(schema_upgraded, table)
            assert cols_sqlite == cols_schema

    def test_copied_values(self, schema_ref, schema_upgraded):
        cursor_ref = get_cursor_for_schema(schema_ref)
        cursor_new = get_cursor_for_schema(schema_upgraded)
        for src_tbl, src_col, dst_tbl, dst_col in self.migration_map:
            # use settings are tested seperately
            if f'use_{dst_tbl}' in get_columns_from_schema(schema_upgraded, 'model_settings'):
                continue
            if dst_col.startswith('use_'):
                continue
            values_ref = get_values_from_sqlite(cursor_ref, src_tbl, src_col)
            values_new = get_values_from_sqlite(cursor_new, dst_tbl, dst_col)
            # dem_file should be different
            if src_col == 'dem_file':
                path_ref = Path(values_ref[0][0])
                assert str(path_ref.name) == values_new[0][0]
            # flow_variable should be different
            elif src_col == 'flow_variable':
                continue
            else:
                assert values_ref == values_new

    @pytest.mark.skip(reason="This test is broken by upgrade to 224")
    def test_boolean_setting_id(self, schema_ref, schema_upgraded):
        cursor_ref = get_cursor_for_schema(schema_ref)
        cursor_new = get_cursor_for_schema(schema_upgraded)
        for col, id, table in self.bool_settings_id:
            id_val = get_values_from_sqlite(cursor_ref, 'v2_global_settings', id)[0][0]
            use_val = get_values_from_sqlite(cursor_new, 'model_settings', col)[0][0]
            settings = get_values_from_sqlite(cursor_new, table, 'id')
            # check if `use_` columns are set properly
            if id_val is None:
                assert (use_val is None or use_val == 0)
            else:
                assert use_val == 1
            # check if matching settings tables consist of 1 (use = True) or 0 (use = False) rows
            if use_val == 0 or use_val is None:
                assert settings == []
            if use_val == 1:
                assert len(settings) == 1

    def test_boolean_setting_exist(self, schema_upgraded):
        cursor_new = get_cursor_for_schema(schema_upgraded)
        for col, table in self.bool_settings_exist:
            use_val = get_values_from_sqlite(cursor_new, 'model_settings', col)[0][0]
            settings = get_values_from_sqlite(cursor_new, table, 'id')
            if use_val == 0 or use_val is None:
                assert settings == []
            if use_val == 1:
                assert len(settings) == 1

    def test_column_length(self, schema_upgraded):
        cursor_new = get_cursor_for_schema(schema_upgraded)
        for table in self.single_row_tables:
            settings = get_values_from_sqlite(cursor_new, table, 'id')
            assert len(settings) == 1



import os
import pathlib
import shutil

import pytest

from threedi_schema import ThreediDatabase

data_dir = pathlib.Path(__file__).parent / "data"


def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    os.environ["SQLALCHEMY_SILENCE_UBER_WARNING"] = "1"


@pytest.fixture
def empty_sqlite_v3(tmp_path):
    """A function-scoped empty spatialite v3 in the latest migration state"""
    tmp_sqlite = tmp_path / "empty_v3.sqlite"
    shutil.copyfile(data_dir / "empty_v3.sqlite", tmp_sqlite)
    return ThreediDatabase(tmp_sqlite)


@pytest.fixture
def empty_sqlite_v4(tmp_path):
    """An function-scoped empty spatialite v4 in the latest migration state"""
    tmp_sqlite = tmp_path / "empty_v4.sqlite"
    shutil.copyfile(data_dir / "empty_v4.sqlite", tmp_sqlite)
    return ThreediDatabase(tmp_sqlite)


@pytest.fixture
def south_latest_sqlite(tmp_path):
    """An empty SQLite that is in its latest South migration state"""
    tmp_sqlite = tmp_path / "south_latest.sqlite"
    shutil.copyfile(data_dir / "south_latest.sqlite", tmp_sqlite)
    return ThreediDatabase(tmp_sqlite)


@pytest.fixture
def oldest_sqlite(tmp_path):
    """A real SQLite that is in its oldest possible south migration state (160)"""
    tmp_sqlite = tmp_path / "noordpolder.sqlite"
    shutil.copyfile(data_dir / "noordpolder.sqlite", tmp_sqlite)
    return ThreediDatabase(tmp_sqlite)


@pytest.fixture
def in_memory_sqlite():
    """An in-memory database with no schema"""
    return ThreediDatabase("")


@pytest.fixture
def sqlite_latest(in_memory_sqlite):
    """An in-memory database with the latest schema version"""
    db = ThreediDatabase("")
    in_memory_sqlite.schema.upgrade("head", backup=False)
    return db



"""Create all tables if they do not exist already.

Revision ID: 0200
Revises:
Create Date: 2021-02-15 16:31:00.792077

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy import inspect, text

from threedi_schema.domain.custom_types import Geometry

# revision identifiers, used by Alembic.
revision = "0200"
down_revision = None
branch_labels = None
depends_on = None


existing_tables = []


def _get_existing_tables(inspector):
    """Fill the global 'existing_tables'"""
    global existing_tables

    existing_tables = inspector.get_table_names()


def create_table_if_not_exists(table_name, *args, **kwargs):
    """Create a table if it is not in the global 'existing_tables'"""
    if table_name in existing_tables:
        return
    else:
        existing_tables.append(table_name)
        return op.create_table(table_name, *args, **kwargs)


def _get_version(connection):
    if "south_migrationhistory" not in existing_tables:
        return
    res = connection.execute(
        text("SELECT id FROM south_migrationhistory ORDER BY id DESC LIMIT 1")
    )
    results = res.fetchall()
    if len(results) == 1:
        return results[0][0]


def upgrade_160():
    """This implements a migration from the old stack:

    0160_auto__add_field_v2controlpid_target_upper_limit__add_field_v2controlpi
    """
    with op.batch_alter_table("v2_control_pid") as batch_op:
        batch_op.add_column(
            sa.Column("target_upper_limit", sa.String(length=50), nullable=True)
        )
        batch_op.add_column(
            sa.Column("target_lower_limit", sa.String(length=50), nullable=True)
        )


def upgrade_161():
    """This implements a migration from the old stack:

    0161_auto__del_field_v2globalsettings_connected_advise_file
    """
    with op.batch_alter_table("v2_global_settings") as batch_op:
        batch_op.drop_column("connected_advise_file")


def upgrade_162():
    """This implements a migration from the old stack:

    0162_auto__add_field_v2globalsettings_table_step_size_1d__add_field_v2globa
    """
    with op.batch_alter_table("v2_global_settings") as batch_op:
        batch_op.add_column(sa.Column("table_step_size_1d", sa.Float(), nullable=True))
        batch_op.add_column(
            sa.Column("table_step_size_volume_2d", sa.Float(), nullable=True)
        )


def upgrade_163():
    """This implements a migration from the old stack:

    0163_auto__add_field_v2globalsettings_use_2d_rain
    """
    with op.batch_alter_table("v2_global_settings") as batch_op:
        batch_op.add_column(sa.Column("use_2d_rain", sa.Integer(), nullable=True))

    op.execute(text("UPDATE v2_global_settings SET use_2d_rain = 1"))


def upgrade_164():
    """This implements a migration from the old stack:

    0164_auto__add_v2gridrefinementarea
    """
    create_table_if_not_exists(
        "v2_grid_refinement_area",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("refinement_level", sa.Integer(), nullable=True),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POLYGON"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )


def upgrade_165():
    """This implements a migration from the old stack:

    0165_auto__add_v2groundwater__add_v2simpleinfiltration__add_v2interflow__ad
    """
    create_table_if_not_exists(
        "v2_groundwater",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("groundwater_impervious_layer_level", sa.Float(), nullable=True),
        sa.Column(
            "groundwater_impervious_layer_level_file",
            sa.String(length=255),
            nullable=True
        ),
        sa.Column(
            "groundwater_impervious_layer_level_type", sa.Integer(), nullable=True
        ),
        sa.Column("phreatic_storage_capacity", sa.Float(), nullable=True),
        sa.Column(
            "phreatic_storage_capacity_file", sa.String(length=255), nullable=True
        ),
        sa.Column("phreatic_storage_capacity_type", sa.Integer(), nullable=True),
        sa.Column("equilibrium_infiltration_rate", sa.Float(), nullable=True),
        sa.Column(
            "equilibrium_infiltration_rate_file", sa.String(length=255), nullable=True
        ),
        sa.Column("equilibrium_infiltration_rate_type", sa.Integer(), nullable=True),
        sa.Column("initial_infiltration_rate", sa.Float(), nullable=True),
        sa.Column(
            "initial_infiltration_rate_file", sa.String(length=255), nullable=True
        ),
        sa.Column("initial_infiltration_rate_type", sa.Integer(), nullable=True),
        sa.Column("infiltration_decay_period", sa.Float(), nullable=True),
        sa.Column(
            "infiltration_decay_period_file", sa.String(length=255), nullable=True
        ),
        sa.Column("infiltration_decay_period_type", sa.Integer(), nullable=True),
        sa.Column("groundwater_hydro_connectivity", sa.Float(), nullable=True),
        sa.Column(
            "groundwater_hydro_connectivity_file", sa.String(length=255), nullable=True
        ),
        sa.Column("groundwater_hydro_connectivity_type", sa.Integer(), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("seepage", sa.Float(), nullable=True),
        sa.Column("seepage_file", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )

    create_table_if_not_exists(
        "v2_simple_infiltration",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("infiltration_rate", sa.Float(), nullable=True),
        sa.Column("infiltration_rate_file", sa.String(length=255), nullable=True),
        sa.Column("infiltration_surface_option", sa.Integer(), nullable=True),
        sa.Column("max_infiltration_capacity_file", sa.Text(), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )

    create_table_if_not_exists(
        "v2_interflow",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("interflow_type", sa.Integer(), nullable=True),
        sa.Column("porosity", sa.Float(), nullable=True),
        sa.Column("porosity_file", sa.String(length=255), nullable=True),
        sa.Column("porosity_layer_thickness", sa.Float(), nullable=True),
        sa.Column("impervious_layer_elevation", sa.Float(), nullable=True),
        sa.Column("hydraulic_conductivity", sa.Float(), nullable=True),
        sa.Column("hydraulic_conductivity_file", sa.String(length=255), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )

    with op.batch_alter_table("v2_global_settings") as batch_op:
        batch_op.add_column(
            sa.Column("initial_groundwater_level", sa.Float(), nullable=True)
        )
        batch_op.add_column(
            sa.Column(
                "initial_groundwater_level_file", sa.String(length=255), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column("initial_groundwater_level_type", sa.Integer(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("groundwater_settings_id", sa.Integer(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("simple_infiltration_settings_id", sa.Integer(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("interflow_settings_id", sa.Integer(), nullable=True)
        )


def upgrade_166():
    """This implements a migration from the old stack:

    0166_fill_Groundwater_Interflow_SimpleInfiltration
    """
    op.execute(text(
        """
    INSERT INTO v2_interflow (id, interflow_type, porosity, porosity_file, porosity_layer_thickness,
                              impervious_layer_elevation, hydraulic_conductivity, hydraulic_conductivity_file)
    SELECT id, interflow_type, porosity, porosity_file, porosity_layer_thickness,
           impervious_layer_elevation, hydraulic_conductivity, hydraulic_conductivity_file
    FROM v2_global_settings;
    """
    ))
    op.execute(text(
        """
    INSERT INTO v2_simple_infiltration (id, infiltration_rate, infiltration_rate_file,
                                        infiltration_surface_option, max_infiltration_capacity_file)
    SELECT id, infiltration_rate, infiltration_rate_file,
           infiltration_surface_option, max_infiltration_capacity_file
    FROM v2_global_settings;
    """
    ))
    op.execute(text(
        """UPDATE v2_global_settings SET interflow_settings_id = id, simple_infiltration_settings_id = id"""
    ))


def upgrade_167():
    """This implements a migration from the old stack:

    0167_auto__del_field_v2globalsettings_infiltration_rate_file__del_field_v2g
    """
    with op.batch_alter_table("v2_global_settings") as batch_op:
        batch_op.drop_column("infiltration_rate_file")
        batch_op.drop_column("infiltration_surface_option")
        batch_op.drop_column("max_infiltration_capacity_file")
        batch_op.drop_column("porosity_layer_thickness")
        batch_op.drop_column("porosity_file")
        batch_op.drop_column("hydraulic_conductivity_file")
        batch_op.drop_column("interflow_type")
        batch_op.drop_column("porosity")
        batch_op.drop_column("impervious_layer_elevation")
        batch_op.drop_column("infiltration_rate")
        batch_op.drop_column("hydraulic_conductivity")


def upgrade_168():
    """This implements a migration from the old stack:

    0168_auto__del_field_v2groundwater_seepage_file__del_field_v2groundwater_se.py
    """
    # Ignored settings table migrations, this table will be removed anyway

    with op.batch_alter_table("v2_groundwater") as batch_op:
        batch_op.drop_column("seepage_file")
        batch_op.drop_column("seepage")
        batch_op.add_column(sa.Column("leakage", sa.Float(), nullable=True))
        batch_op.add_column(
            sa.Column("leakage_file", sa.String(length=255), nullable=True)
        )


def upgrade_169():
    """This implements a migration from the old stack:

    0169_auto__del_field_mdusettings_leakage_file__add_field_mdusettings_seepag.py
    """
    # Ignored settings table migrations, this table will be removed anyway


def upgrade_170():
    """This implements a migration from the old stack:

    0170_auto__chg_field_v2aggregationsettings_aggregation_method.py
    """
    with op.batch_alter_table("v2_aggregation_settings") as batch_op:
        batch_op.alter_column("aggregation_method", type_=sa.String(length=100))


def upgrade_171():
    """This implements a migration from the old stack:

    0170_auto__chg_field_v2culvert_discharge_coefficient_negative__chg_field_v2.py
    """
    op.execute(text(
        "UPDATE v2_culvert SET discharge_coefficient_negative = 1.0 WHERE discharge_coefficient_negative IS NULL"
    ))
    op.execute(text(
        "UPDATE v2_culvert SET discharge_coefficient_positive = 1.0 WHERE discharge_coefficient_positive IS NULL"
    ))


def upgrade_172():
    """This implements a migration from the old stack:

    0171_auto__chg_field_v2aggregationsettings_aggregation_method__del_field_v2.py
    """
    with op.batch_alter_table("v2_global_settings") as batch_op:
        batch_op.alter_column(
            "max_interception_file", new_column_name="interception_file"
        )
        batch_op.alter_column("max_interception", new_column_name="interception_global")


def upgrade_173():
    """This implements a migration from the old stack:

    0172_auto__del_v2initialwaterlevel__del_field_v2orifice_max_capacity__del_f
    """
    op.drop_table("v2_initial_waterlevel")

    with op.batch_alter_table("v2_orifice") as batch_op:
        batch_op.drop_column("max_capacity")

    with op.batch_alter_table("v2_impervious_surface") as batch_op:
        batch_op.drop_column("function")

    with op.batch_alter_table("v2_pipe") as batch_op:
        batch_op.drop_column("pipe_quality")

    with op.batch_alter_table("v2_culvert") as batch_op:
        batch_op.alter_column("discharge_coefficient_negative", nullable=True)
        batch_op.alter_column("discharge_coefficient_positive", nullable=True)


UPGRADE_LOOKUP = {
    160: upgrade_160,
    161: upgrade_161,
    162: upgrade_162,
    163: upgrade_163,
    164: upgrade_164,
    165: upgrade_165,
    166: upgrade_166,
    167: upgrade_167,
    168: upgrade_168,
    169: upgrade_169,
    170: upgrade_170,
    171: upgrade_171,
    172: upgrade_172,
    173: upgrade_173,
}


def upgrade():
    conn = op.get_bind()
    inspector = inspect(conn)

    # Setup the global 'existing_tables'
    _get_existing_tables(inspector)

    # Initialize the Spatialite if necessary:
    if conn.dialect.name == "sqlite" and "spatial_ref_sys" not in existing_tables:
        # The (1) performs the init in 1 transaction, which improves performance.
        op.execute(text("SELECT InitSpatialMetadata(1)"))

    version = _get_version(conn)
    if version is not None:
        for i in range(version, 174):
            UPGRADE_LOOKUP[i]()

    create_table_if_not_exists(
        "v2_2d_boundary_conditions",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("timeseries", sa.Text(), nullable=True),
        sa.Column("boundary_type", sa.Integer(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "LINESTRING"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_2d_lateral",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("type", sa.Integer(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POINT"
                
                
            ),
            nullable=True
        ),
        sa.Column("timeseries", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_calculation_point",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("content_type_id", sa.Integer(), nullable=True),
        sa.Column("user_ref", sa.String(length=80), nullable=True),
        sa.Column("calc_type", sa.Integer(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POINT"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_connection_nodes",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("storage_area", sa.Float(), nullable=True),
        sa.Column("initial_waterlevel", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POINT"
                
                
            ),
            nullable=True
        ),
        sa.Column(
            "the_geom_linestring",
            Geometry(
                "LINESTRING"
                
                
            ),
            nullable=True
        ),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control_delta",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("measure_variable", sa.String(length=50), nullable=True),
        sa.Column("measure_delta", sa.String(length=50), nullable=True),
        sa.Column("measure_dt", sa.Float(), nullable=True),
        sa.Column("action_type", sa.String(length=50), nullable=True),
        sa.Column("action_value", sa.String(length=50), nullable=True),
        sa.Column("action_time", sa.Float(), nullable=True),
        sa.Column("target_type", sa.String(length=100), nullable=True),
        sa.Column("target_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control_group",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control_measure_group",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control_memory",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("measure_variable", sa.String(length=50), nullable=True),
        sa.Column("upper_threshold", sa.Float(), nullable=True),
        sa.Column("lower_threshold", sa.Float(), nullable=True),
        sa.Column("action_type", sa.String(length=50), nullable=True),
        sa.Column("action_value", sa.String(length=50), nullable=True),
        sa.Column("target_type", sa.String(length=100), nullable=True),
        sa.Column("target_id", sa.Integer(), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.Column("is_inverse", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control_pid",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("measure_variable", sa.String(length=50), nullable=True),
        sa.Column("setpoint", sa.Float(), nullable=True),
        sa.Column("kp", sa.Float(), nullable=True),
        sa.Column("ki", sa.Float(), nullable=True),
        sa.Column("kd", sa.Float(), nullable=True),
        sa.Column("action_type", sa.String(length=50), nullable=True),
        sa.Column("target_type", sa.String(length=100), nullable=True),
        sa.Column("target_upper_limit", sa.String(length=50), nullable=True),
        sa.Column("target_lower_limit", sa.String(length=50), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control_table",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("action_table", sa.Text(), nullable=True),
        sa.Column("action_type", sa.String(length=50), nullable=True),
        sa.Column("measure_variable", sa.String(length=50), nullable=True),
        sa.Column("measure_operator", sa.String(length=2), nullable=True),
        sa.Column("target_type", sa.String(length=100), nullable=True),
        sa.Column("target_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control_timed",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("action_type", sa.String(length=50), nullable=True),
        sa.Column("action_table", sa.Text(), nullable=True),
        sa.Column("target_type", sa.String(length=100), nullable=True),
        sa.Column("target_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_cross_section_definition",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("width", sa.String(length=255), nullable=True),
        sa.Column("height", sa.String(length=255), nullable=True),
        sa.Column("shape", sa.Integer(), nullable=True),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_dem_average_area",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(
            "the_geom",
            Geometry(
                "POLYGON"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_floodfill",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("waterlevel", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POINT"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_grid_refinement",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("refinement_level", sa.Integer(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "LINESTRING"
                
                
            ),
            nullable=True
        ),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_grid_refinement_area",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("refinement_level", sa.Integer(), nullable=True),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POLYGON"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_groundwater",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("groundwater_impervious_layer_level", sa.Float(), nullable=True),
        sa.Column(
            "groundwater_impervious_layer_level_file",
            sa.String(length=255),
            nullable=True
        ),
        sa.Column(
            "groundwater_impervious_layer_level_type", sa.Integer(), nullable=True
        ),
        sa.Column("phreatic_storage_capacity", sa.Float(), nullable=True),
        sa.Column(
            "phreatic_storage_capacity_file", sa.String(length=255), nullable=True
        ),
        sa.Column("phreatic_storage_capacity_type", sa.Integer(), nullable=True),
        sa.Column("equilibrium_infiltration_rate", sa.Float(), nullable=True),
        sa.Column(
            "equilibrium_infiltration_rate_file", sa.String(length=255), nullable=True
        ),
        sa.Column("equilibrium_infiltration_rate_type", sa.Integer(), nullable=True),
        sa.Column("initial_infiltration_rate", sa.Float(), nullable=True),
        sa.Column(
            "initial_infiltration_rate_file", sa.String(length=255), nullable=True
        ),
        sa.Column("initial_infiltration_rate_type", sa.Integer(), nullable=True),
        sa.Column("infiltration_decay_period", sa.Float(), nullable=True),
        sa.Column(
            "infiltration_decay_period_file", sa.String(length=255), nullable=True
        ),
        sa.Column("infiltration_decay_period_type", sa.Integer(), nullable=True),
        sa.Column("groundwater_hydro_connectivity", sa.Float(), nullable=True),
        sa.Column(
            "groundwater_hydro_connectivity_file", sa.String(length=255), nullable=True
        ),
        sa.Column("groundwater_hydro_connectivity_type", sa.Integer(), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("leakage", sa.Float(), nullable=True),
        sa.Column("leakage_file", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_impervious_surface",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("surface_inclination", sa.String(length=64), nullable=True),
        sa.Column("surface_class", sa.String(length=128), nullable=True),
        sa.Column("surface_sub_class", sa.String(length=128), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column("nr_of_inhabitants", sa.Float(), nullable=True),
        sa.Column("area", sa.Float(), nullable=True),
        sa.Column("dry_weather_flow", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POLYGON"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_interflow",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("interflow_type", sa.Integer(), nullable=True),
        sa.Column("porosity", sa.Float(), nullable=True),
        sa.Column("porosity_file", sa.String(length=255), nullable=True),
        sa.Column("porosity_layer_thickness", sa.Float(), nullable=True),
        sa.Column("impervious_layer_elevation", sa.Float(), nullable=True),
        sa.Column("hydraulic_conductivity", sa.Float(), nullable=True),
        sa.Column("hydraulic_conductivity_file", sa.String(length=255), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_levee",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("crest_level", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "LINESTRING"
                
                
            ),
            nullable=True
        ),
        sa.Column("material", sa.Integer(), nullable=True),
        sa.Column("max_breach_depth", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_numerical_settings",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("cfl_strictness_factor_1d", sa.Float(), nullable=True),
        sa.Column("cfl_strictness_factor_2d", sa.Float(), nullable=True),
        sa.Column("convergence_cg", sa.Float(), nullable=True),
        sa.Column("convergence_eps", sa.Float(), nullable=True),
        sa.Column("flow_direction_threshold", sa.Float(), nullable=True),
        sa.Column("frict_shallow_water_correction", sa.Integer(), nullable=True),
        sa.Column("general_numerical_threshold", sa.Float(), nullable=True),
        sa.Column("integration_method", sa.Integer(), nullable=True),
        sa.Column("limiter_grad_1d", sa.Integer(), nullable=True),
        sa.Column("limiter_grad_2d", sa.Integer(), nullable=True),
        sa.Column("limiter_slope_crossectional_area_2d", sa.Integer(), nullable=True),
        sa.Column("limiter_slope_friction_2d", sa.Integer(), nullable=True),
        sa.Column("max_nonlin_iterations", sa.Integer(), nullable=True),
        sa.Column("max_degree", sa.Integer(), nullable=True),
        sa.Column("minimum_friction_velocity", sa.Float(), nullable=True),
        sa.Column("minimum_surface_area", sa.Float(), nullable=True),
        sa.Column("precon_cg", sa.Integer(), nullable=True),
        sa.Column("preissmann_slot", sa.Float(), nullable=True),
        sa.Column("pump_implicit_ratio", sa.Float(), nullable=True),
        sa.Column("thin_water_layer_definition", sa.Float(), nullable=True),
        sa.Column("use_of_cg", sa.Integer(), nullable=True),
        sa.Column("use_of_nested_newton", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_obstacle",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("crest_level", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "LINESTRING"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_simple_infiltration",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("infiltration_rate", sa.Float(), nullable=True),
        sa.Column("infiltration_rate_file", sa.String(length=255), nullable=True),
        sa.Column("infiltration_surface_option", sa.Integer(), nullable=True),
        sa.Column("max_infiltration_capacity_file", sa.Text(), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_surface_parameters",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("outflow_delay", sa.Float(), nullable=True),
        sa.Column("surface_layer_thickness", sa.Float(), nullable=True),
        sa.Column("infiltration", sa.Boolean(), nullable=True),
        sa.Column("max_infiltration_capacity", sa.Float(), nullable=True),
        sa.Column("min_infiltration_capacity", sa.Float(), nullable=True),
        sa.Column("infiltration_decay_constant", sa.Float(), nullable=True),
        sa.Column("infiltration_recovery_constant", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_1d_boundary_conditions",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("boundary_type", sa.Integer(), nullable=True),
        sa.Column("timeseries", sa.Text(), nullable=True),
        sa.Column("connection_node_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_1d_lateral",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("connection_node_id", sa.Integer(), nullable=True),
        sa.Column("timeseries", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_channel",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("calculation_type", sa.Integer(), nullable=True),
        sa.Column("dist_calc_points", sa.Float(), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "LINESTRING"
                
                
            ),
            nullable=True
        ),
        sa.Column("connection_node_start_id", sa.Integer(), nullable=True),
        sa.Column("connection_node_end_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_connected_pnt",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("calculation_pnt_id", sa.Integer(), nullable=True),
        sa.Column("levee_id", sa.Integer(), nullable=True),
        sa.Column("exchange_level", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POINT"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("control_group_id", sa.Integer(), nullable=True),
        sa.Column("measure_group_id", sa.Integer(), nullable=True),
        sa.Column("control_type", sa.String(length=15), nullable=True),
        sa.Column("control_id", sa.Integer(), nullable=True),
        sa.Column("start", sa.String(length=50), nullable=True),
        sa.Column("end", sa.String(length=50), nullable=True),
        sa.Column("measure_frequency", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_control_measure_map",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("measure_group_id", sa.Integer(), nullable=True),
        sa.Column("object_type", sa.String(length=100), nullable=True),
        sa.Column("object_id", sa.Integer(), nullable=True),
        sa.Column("weight", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_culvert",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("calculation_type", sa.Integer(), nullable=True),
        sa.Column("friction_value", sa.Float(), nullable=True),
        sa.Column("friction_type", sa.Integer(), nullable=True),
        sa.Column("dist_calc_points", sa.Float(), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column("discharge_coefficient_positive", sa.Float(), nullable=True),
        sa.Column("discharge_coefficient_negative", sa.Float(), nullable=True),
        sa.Column("invert_level_start_point", sa.Float(), nullable=True),
        sa.Column("invert_level_end_point", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "LINESTRING"
                
                
            ),
            nullable=True
        ),
        sa.Column("connection_node_start_id", sa.Integer(), nullable=True),
        sa.Column("connection_node_end_id", sa.Integer(), nullable=True),
        sa.Column("cross_section_definition_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_global_settings",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("use_2d_flow", sa.Boolean(), nullable=True),
        sa.Column("use_1d_flow", sa.Boolean(), nullable=True),
        sa.Column("manhole_storage_area", sa.Float(), nullable=True),
        sa.Column("name", sa.String(length=128), nullable=True),
        sa.Column("sim_time_step", sa.Float(), nullable=True),
        sa.Column("output_time_step", sa.Float(), nullable=True),
        sa.Column("nr_timesteps", sa.Integer(), nullable=True),
        sa.Column("start_time", sa.Text(), nullable=True),
        sa.Column("start_date", sa.Text(), nullable=True),
        sa.Column("grid_space", sa.Float(), nullable=True),
        sa.Column("dist_calc_points", sa.Float(), nullable=True),
        sa.Column("kmax", sa.Integer(), nullable=True),
        sa.Column("guess_dams", sa.Integer(), nullable=True),
        sa.Column("table_step_size", sa.Float(), nullable=True),
        sa.Column("flooding_threshold", sa.Float(), nullable=True),
        sa.Column("advection_1d", sa.Integer(), nullable=True),
        sa.Column("advection_2d", sa.Integer(), nullable=True),
        sa.Column("dem_file", sa.String(length=255), nullable=True),
        sa.Column("frict_type", sa.Integer(), nullable=True),
        sa.Column("frict_coef", sa.Float(), nullable=True),
        sa.Column("frict_coef_file", sa.String(length=255), nullable=True),
        sa.Column("water_level_ini_type", sa.Integer(), nullable=True),
        sa.Column("initial_waterlevel", sa.Float(), nullable=True),
        sa.Column("initial_waterlevel_file", sa.String(length=255), nullable=True),
        sa.Column("interception_global", sa.Float(), nullable=True),
        sa.Column("interception_file", sa.String(length=255), nullable=True),
        sa.Column("dem_obstacle_detection", sa.Boolean(), nullable=True),
        sa.Column("dem_obstacle_height", sa.Float(), nullable=True),
        sa.Column("embedded_cutoff_threshold", sa.Float(), nullable=True),
        sa.Column("epsg_code", sa.Integer(), nullable=True),
        sa.Column("timestep_plus", sa.Boolean(), nullable=True),
        sa.Column("max_angle_1d_advection", sa.Float(), nullable=True),
        sa.Column("minimum_sim_time_step", sa.Float(), nullable=True),
        sa.Column("maximum_sim_time_step", sa.Float(), nullable=True),
        sa.Column("frict_avg", sa.Integer(), nullable=True),
        sa.Column("wind_shielding_file", sa.String(length=255), nullable=True),
        sa.Column("use_0d_inflow", sa.Integer(), nullable=True),
        sa.Column("table_step_size_1d", sa.Float(), nullable=True),
        sa.Column("table_step_size_volume_2d", sa.Float(), nullable=True),
        sa.Column("use_2d_rain", sa.Integer(), nullable=True),
        sa.Column("initial_groundwater_level", sa.Float(), nullable=True),
        sa.Column(
            "initial_groundwater_level_file", sa.String(length=255), nullable=True
        ),
        sa.Column("initial_groundwater_level_type", sa.Integer(), nullable=True),
        sa.Column("numerical_settings_id", sa.Integer(), nullable=True),
        sa.Column("interflow_settings_id", sa.Integer(), nullable=True),
        sa.Column("control_group_id", sa.Integer(), nullable=True),
        sa.Column("simple_infiltration_settings_id", sa.Integer(), nullable=True),
        sa.Column("groundwater_settings_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_impervious_surface_map",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("percentage", sa.Float(), nullable=True),
        sa.Column("impervious_surface_id", sa.Integer(), nullable=True),
        sa.Column("connection_node_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_manhole",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column("shape", sa.String(length=4), nullable=True),
        sa.Column("width", sa.Float(), nullable=True),
        sa.Column("length", sa.Float(), nullable=True),
        sa.Column("surface_level", sa.Float(), nullable=True),
        sa.Column("bottom_level", sa.Float(), nullable=True),
        sa.Column("drain_level", sa.Float(), nullable=True),
        sa.Column("sediment_level", sa.Float(), nullable=True),
        sa.Column("manhole_indicator", sa.Integer(), nullable=True),
        sa.Column("calculation_type", sa.Integer(), nullable=True),
        sa.Column("connection_node_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_orifice",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column("crest_type", sa.Integer(), nullable=True),
        sa.Column("crest_level", sa.Float(), nullable=True),
        sa.Column("friction_value", sa.Float(), nullable=True),
        sa.Column("friction_type", sa.Integer(), nullable=True),
        sa.Column("discharge_coefficient_positive", sa.Float(), nullable=True),
        sa.Column("discharge_coefficient_negative", sa.Float(), nullable=True),
        sa.Column("sewerage", sa.Boolean(), nullable=True),
        sa.Column("connection_node_start_id", sa.Integer(), nullable=True),
        sa.Column("connection_node_end_id", sa.Integer(), nullable=True),
        sa.Column("cross_section_definition_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_pipe",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("profile_num", sa.Integer(), nullable=True),
        sa.Column("sewerage_type", sa.Integer(), nullable=True),
        sa.Column("calculation_type", sa.Integer(), nullable=True),
        sa.Column("invert_level_start_point", sa.Float(), nullable=True),
        sa.Column("invert_level_end_point", sa.Float(), nullable=True),
        sa.Column("friction_value", sa.Float(), nullable=True),
        sa.Column("friction_type", sa.Integer(), nullable=True),
        sa.Column("dist_calc_points", sa.Float(), nullable=True),
        sa.Column("material", sa.Integer(), nullable=True),
        sa.Column("original_length", sa.Float(), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column("connection_node_start_id", sa.Integer(), nullable=True),
        sa.Column("connection_node_end_id", sa.Integer(), nullable=True),
        sa.Column("cross_section_definition_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_pumpstation",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column("classification", sa.Integer(), nullable=True),
        sa.Column("sewerage", sa.Boolean(), nullable=True),
        sa.Column("type", sa.Integer(), nullable=True),
        sa.Column("start_level", sa.Float(), nullable=True),
        sa.Column("lower_stop_level", sa.Float(), nullable=True),
        sa.Column("upper_stop_level", sa.Float(), nullable=True),
        sa.Column("capacity", sa.Float(), nullable=True),
        sa.Column("connection_node_start_id", sa.Integer(), nullable=True),
        sa.Column("connection_node_end_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_surface",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column("nr_of_inhabitants", sa.Float(), nullable=True),
        sa.Column("dry_weather_flow", sa.Float(), nullable=True),
        sa.Column("function", sa.String(length=64), nullable=True),
        sa.Column("area", sa.Float(), nullable=True),
        sa.Column("surface_parameters_id", sa.Integer(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POLYGON"
                
                
            ),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_surface_map",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("surface_type", sa.String(length=40), nullable=True),
        sa.Column("surface_id", sa.Integer(), nullable=True),
        sa.Column("connection_node_id", sa.Integer(), nullable=True),
        sa.Column("percentage", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_weir",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("crest_level", sa.Float(), nullable=True),
        sa.Column("crest_type", sa.Integer(), nullable=True),
        sa.Column("friction_value", sa.Float(), nullable=True),
        sa.Column("friction_type", sa.Integer(), nullable=True),
        sa.Column("discharge_coefficient_positive", sa.Float(), nullable=True),
        sa.Column("discharge_coefficient_negative", sa.Float(), nullable=True),
        sa.Column("sewerage", sa.Boolean(), nullable=True),
        sa.Column("external", sa.Boolean(), nullable=True),
        sa.Column("zoom_category", sa.Integer(), nullable=True),
        sa.Column("connection_node_start_id", sa.Integer(), nullable=True),
        sa.Column("connection_node_end_id", sa.Integer(), nullable=True),
        sa.Column("cross_section_definition_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_aggregation_settings",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("global_settings_id", sa.Integer(), nullable=True),
        sa.Column("var_name", sa.String(length=100), nullable=True),
        sa.Column("flow_variable", sa.String(length=100), nullable=True),
        sa.Column("aggregation_method", sa.String(length=100), nullable=True),
        sa.Column("aggregation_in_space", sa.Boolean(), nullable=True),
        sa.Column("timestep", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_cross_section_location",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("reference_level", sa.Float(), nullable=True),
        sa.Column("friction_type", sa.Integer(), nullable=True),
        sa.Column("friction_value", sa.Float(), nullable=True),
        sa.Column("bank_level", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POINT"
                
                
            ),
            nullable=True
        ),
        sa.Column("channel_id", sa.Integer(), nullable=True),
        sa.Column("definition_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    create_table_if_not_exists(
        "v2_windshielding",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("north", sa.Float(), nullable=True),
        sa.Column("northeast", sa.Float(), nullable=True),
        sa.Column("east", sa.Float(), nullable=True),
        sa.Column("southeast", sa.Float(), nullable=True),
        sa.Column("south", sa.Float(), nullable=True),
        sa.Column("southwest", sa.Float(), nullable=True),
        sa.Column("west", sa.Float(), nullable=True),
        sa.Column("northwest", sa.Float(), nullable=True),
        sa.Column(
            "the_geom",
            Geometry(
                "POINT"
                
                
            ),
            nullable=True
        ),
        sa.Column("channel_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    pass



"""Upgrade settings in schema

Revision ID: 0225
Revises:
Create Date: 2024-09-10 09:00

"""
import csv
import uuid
from pathlib import Path
from typing import Dict, List, Tuple

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Column, Float, func, Integer, select, String
from sqlalchemy.orm import declarative_base, Session

from threedi_schema.domain import constants, models
from threedi_schema.domain.custom_types import IntegerEnum

Base = declarative_base()

data_dir = Path(__file__).parent / "data"


# revision identifiers, used by Alembic.
revision = "0228"
down_revision = "0227"
branch_labels = None
depends_on = None

RENAME_TABLES = [
    ("v2_channel", "channel"),
    ("v2_windshielding", "windshielding_1d"),
    ("v2_cross_section_location", "cross_section_location"),
    ("v2_pipe", "pipe"),
    ("v2_culvert", "culvert"),
    ("v2_weir", "weir"),
    ("v2_orifice", "orifice"),
    ("v2_pumpstation", "pump")
]

DELETE_TABLES = ["v2_cross_section_definition",
                 "v2_floodfill",
                 "v2_connection_nodes"]

RENAME_COLUMNS = {
    "culvert": {"calculation_type": "exchange_type",
                "dist_calc_points": "calculation_point_distance",
                "invert_level_start_point": "invert_level_start",
                "invert_level_end_point": "invert_level_end",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "pipe": {"calculation_type": "exchange_type",
             "dist_calc_points": "calculation_point_distance",
             "material": "material_id",
                "invert_level_start_point": "invert_level_start",
                "invert_level_end_point": "invert_level_end",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "channel": {"calculation_type": "exchange_type",
                "dist_calc_points": "calculation_point_distance",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "weir": {"calculation_type": "exchange_type",
                "dist_calc_points": "calculation_point_distance",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "orifice": {"calculation_type": "exchange_type",
                "dist_calc_points": "calculation_point_distance",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "pump": {"connection_node_start_id": "connection_node_id"}
}

REMOVE_COLUMNS = {
    "channel": ["zoom_category",],
    "cross_section_location": ["definition_id", "vegetation_drag_coeficients"],
    "culvert": ["zoom_category", "cross_section_definition_id"],
    "pipe": ["zoom_category", "original_length", "cross_section_definition_id", "profile_num"],
    "orifice": ["zoom_category", "cross_section_definition_id"],
    "weir": ["zoom_category", "cross_section_definition_id"],
    "pump": ["connection_node_end_id", "zoom_category", "classification"]
}

RETYPE_COLUMNS = {}


class Schema228UpgradeException(Exception):
    pass


def add_columns_to_tables(table_columns: List[Tuple[str, Column]]):
    # no checks for existence are done, this will fail if any column already exists
    for dst_table, col in table_columns:
        with op.batch_alter_table(dst_table) as batch_op:
            batch_op.add_column(col)


def remove_tables(tables: List[str]):
    for table in tables:
        op.drop_table(table)


def modify_table(old_table_name, new_table_name):
    # Create a new table named `new_table_name` using the declared models
    # Use the columns from `old_table_name`, with the following exceptions:
    # * columns in `RENAME_COLUMNS[new_table_name]` are renamed
    # * `the_geom` is renamed to `geom` and NOT NULL is enforced
    model = find_model(new_table_name)
    # create new table
    create_sqlite_table_from_model(model)
    # get column names from model and match them to available data in sqlite
    connection = op.get_bind()
    rename_cols = {**RENAME_COLUMNS.get(new_table_name, {}), "the_geom": "geom"}
    rename_cols_rev = {v: k for k, v in rename_cols.items()}
    col_map = [(col.name, rename_cols_rev.get(col.name, col.name)) for col in get_cols_for_model(model)]
    available_cols = [col[1] for col in connection.execute(sa.text(f"PRAGMA table_info('{old_table_name}')")).fetchall()]
    new_col_names, old_col_names = zip(*[(new_col, old_col) for new_col, old_col in col_map if old_col in available_cols])
    # Copy data
    # This may copy wrong type data because some types change!!
    op.execute(sa.text(f"INSERT INTO {new_table_name} ({','.join(new_col_names)}) "
                       f"SELECT {','.join(old_col_names)} FROM {old_table_name}"))


def find_model(table_name):
    for model in models.DECLARED_MODELS:
        if model.__tablename__ == table_name:
            return model
    # This can only go wrong if the migration or model is incorrect
    raise

def fix_geometry_columns():
    update_models = [models.Channel, models.ConnectionNode, models.CrossSectionLocation,
                     models.Culvert, models.Orifice, models.Pipe, models.Pump,
                     models.PumpMap, models.Weir, models.Windshielding]
    for model in update_models:
        op.execute(sa.text(f"SELECT RecoverGeometryColumn('{model.__tablename__}', "
                           f"'geom', {4326}, '{model.geom.type.geometry_type}', 'XY')"))
        op.execute(sa.text(f"SELECT CreateSpatialIndex('{model.__tablename__}', 'geom')"))


class Temp(Base):
    __tablename__ = f'_temp_228_{uuid.uuid4().hex}'

    id = Column(Integer, primary_key=True)
    cross_section_table = Column(String)
    cross_section_friction_values = Column(String)
    cross_section_vegetation_table = Column(String)
    cross_section_shape = Column(IntegerEnum(constants.CrossSectionShape))


def extend_cross_section_definition_table():
    conn = op.get_bind()
    session = Session(bind=op.get_bind())
    # create temporary table
    op.execute(sa.text(
        f"""CREATE TABLE {Temp.__tablename__} 
            (id INTEGER PRIMARY KEY, 
             cross_section_table TEXT,
             cross_section_shape INT,
             cross_section_width REAL,
             cross_section_height REAL,
             cross_section_friction_values TEXT,
             cross_section_vegetation_table TEXT)
        """))
    # copy id's from v2_cross_section_definition
    op.execute(sa.text(
        f"""INSERT INTO {Temp.__tablename__} (id, cross_section_shape, cross_section_width, cross_section_height) 
           SELECT id, shape, width, height 
           FROM v2_cross_section_definition"""
    ))
    for col_name in ['cross_section_width', 'cross_section_height']:
        op.execute(sa.text(f"""
                UPDATE {Temp.__tablename__}
                SET {col_name} = NULL 
                WHERE {col_name} = '';
            """))

    def make_table(*args):
        split_args = [arg.split() for arg in args]
        if not all(len(args) == len(split_args[0]) for args in split_args):
            return
        return '\n'.join([','.join(row) for row in zip(*split_args)])
    # Create cross_section_table for tabulated
    res = conn.execute(sa.text(f"""
        SELECT id, height, width, shape FROM v2_cross_section_definition 
        WHERE v2_cross_section_definition.shape IN (5,6,7)   
        AND height IS NOT NULL AND width IS NOT NULL
    """)).fetchall()
    for id, h, w, s in res:
        temp_row = session.query(Temp).filter_by(id=id).first()
        # tabulated_YZ: width -> Y; height -> Z
        if s == constants.CrossSectionShape.TABULATED_YZ.value:
            temp_row.cross_section_table = make_table(w, h)
        # tabulated_trapezium or tabulated_rectangle: height, width
        else:
            temp_row.cross_section_table = make_table(h, w)
        session.commit()
    # add cross_section_friction_table to cross_section_definition
    res = conn.execute(sa.text("""
        SELECT id, friction_values FROM v2_cross_section_definition 
        WHERE friction_values IS NOT NULL
        AND v2_cross_section_definition.shape = 7 
    """)).fetchall()
    for id, friction_values in res:
        temp_row = session.query(Temp).filter_by(id=id).first()
        temp_row.cross_section_friction_values = friction_values.replace(' ',',')
        session.commit()
    # add cross_section_vegetation_table to cross_section_definition
    res = conn.execute(sa.text("""
        SELECT id, vegetation_stem_densities, vegetation_stem_diameters, vegetation_heights, vegetation_drag_coefficients
        FROM v2_cross_section_definition 
        WHERE vegetation_stem_densities IS NOT NULL
        AND vegetation_stem_diameters IS NOT NULL
        AND vegetation_heights IS NOT NULL
        AND v2_cross_section_definition.shape = 7 
        AND vegetation_drag_coefficients IS NOT NULL
    """)).fetchall()
    for id, dens, diam, h, c in res:
        temp_row = session.query(Temp).filter_by(id=id).first()
        temp_row.cross_section_vegetation_table = make_table(dens, diam, h, c)
        session.commit()


def migrate_cross_section_definition_from_temp(target_table: str,
                                               cols: List[Tuple[str, str]],
                                               def_id_col: str):
    for cname, ctype in cols:
        op.execute(sa.text(f'ALTER TABLE {target_table} ADD COLUMN {cname} {ctype}'))
    # ensure that types work properly
    # e.g. heights cannot be text!!
    set_query = ','.join(
        f'{cname} = (SELECT {cname} FROM {Temp.__tablename__} WHERE id = {target_table}.{def_id_col})' for cname, _ in
        cols)
    op.execute(sa.text(f"""
        UPDATE {target_table}
        SET {set_query}
        WHERE EXISTS (SELECT 1 FROM {Temp.__tablename__} WHERE id = {target_table}.{def_id_col});
    """))
    op.execute(sa.text(f"UPDATE {target_table} SET cross_section_width = NULL WHERE cross_section_shape IN (5,6,7)"))
    op.execute(sa.text(f"UPDATE {target_table} SET cross_section_height = NULL WHERE cross_section_shape IN (5,6,7)"))



def migrate_cross_section_definition_to_location():
    cols = [('cross_section_table', 'TEXT'),
            ('cross_section_friction_values', 'TEXT'),
            ('cross_section_vegetation_table', 'TEXT'),
            ('cross_section_shape', 'INT'),
            ('cross_section_width', 'REAL'),
            ('cross_section_height', 'REAL')]
    migrate_cross_section_definition_from_temp(target_table='v2_cross_section_location',
                                               cols=cols,
                                               def_id_col='definition_id')

def migrate_cross_section_definition_to_object(table_name: str):
    cols = [('cross_section_table', 'TEXT'),
            ('cross_section_shape', 'INT'),
            ('cross_section_width', 'REAL'),
            ('cross_section_height', 'REAL')]
    migrate_cross_section_definition_from_temp(target_table=table_name,
                                               cols=cols,
                                               def_id_col='cross_section_definition_id')


def set_geom_for_object(table_name: str, col_name: str = 'the_geom'):
    # line from connection_node_start_id to connection_node_end_id
    # SELECT load_extension('mod_spatialite');
    op.execute(sa.text(f"SELECT AddGeometryColumn('{table_name}', '{col_name}', 4326, 'LINESTRING', 'XY', 0);"))
    q = f"""
        UPDATE
            {table_name}
        SET the_geom = (
            SELECT MakeLine(start_node.the_geom, end_node.the_geom) FROM {table_name} AS object 
            JOIN v2_connection_nodes AS start_node ON object.connection_node_start_id = start_node.id
            JOIN v2_connection_nodes AS end_node ON object.connection_node_end_id = end_node.id
        )         
    """
    op.execute(sa.text(q))


def set_geom_for_v2_pumpstation():
    op.execute(sa.text(f"SELECT AddGeometryColumn('v2_pumpstation', 'the_geom', 4326, 'POINT', 'XY', 0);"))
    q = f"""
        UPDATE
            v2_pumpstation
        SET the_geom = (
            SELECT node.the_geom FROM v2_pumpstation AS object 
            JOIN v2_connection_nodes AS node ON object.connection_node_start_id = node.id
        )         
    """
    op.execute(sa.text(q))


def get_cols_for_model(model, skip_cols=None):
    from sqlalchemy.orm.attributes import InstrumentedAttribute
    if skip_cols is None:
        skip_cols = []
    return [getattr(model, item) for item in model.__dict__
            if item not in skip_cols
            and isinstance(getattr(model, item), InstrumentedAttribute)]


def create_sqlite_table_from_model(model):
    cols = get_cols_for_model(model, skip_cols = ["id", "geom"])
    op.execute(sa.text(f"""
        CREATE TABLE {model.__tablename__} (
        id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, 
        {','.join(f"{col.name} {col.type}" for col in cols)},
        geom {model.geom.type.geometry_type} NOT NULL
    );"""))


def create_pump_map():
    # Create table
    create_sqlite_table_from_model(models.PumpMap)

    # Create geometry
    op.execute(sa.text(f"SELECT AddGeometryColumn('v2_pumpstation', 'map_geom', 4326, 'LINESTRING', 'XY', 0);"))
    op.execute(sa.text("""
        UPDATE v2_pumpstation
        SET map_geom = (
            SELECT MakeLine(start_geom.the_geom, end_geom.the_geom)
            FROM v2_connection_nodes AS start_geom, v2_connection_nodes AS end_geom
            WHERE v2_pumpstation.connection_node_start_id = start_geom.id
            AND v2_pumpstation.connection_node_end_id = end_geom.id
        )
        WHERE EXISTS (
            SELECT 1
            FROM v2_connection_nodes AS start_geom, v2_connection_nodes AS end_geom
            WHERE v2_pumpstation.connection_node_start_id = start_geom.id
            AND v2_pumpstation.connection_node_end_id = end_geom.id
        );
    """))

    # Copy data from v2_pumpstation
    new_col_names = ["pump_id", "connection_node_id_end", "code", "display_name", "geom"]
    old_col_names = ["id", "connection_node_end_id", "code", "display_name", "map_geom"]
    op.execute(sa.text(f"""
        INSERT INTO pump_map ({','.join(new_col_names)}) 
        SELECT {','.join(old_col_names)} FROM v2_pumpstation
        WHERE v2_pumpstation.connection_node_end_id IS NOT NULL
        AND v2_pumpstation.connection_node_start_id IS NOT NULL
    """))


def create_connection_node():
    create_sqlite_table_from_model(models.ConnectionNode)
    # copy from v2_connection_nodes
    old_col_names = ["id", "initial_waterlevel", "storage_area", "the_geom", "code"]
    rename_map = {"initial_waterlevel": "initial_water_level", "the_geom": "geom"}
    new_col_names = [rename_map.get(old_name, old_name) for old_name in old_col_names]
    op.execute(sa.text(f"""
        INSERT INTO connection_node ({','.join(new_col_names)}) 
        SELECT {','.join(old_col_names)} FROM v2_connection_nodes
    """))
    # conditional copy from v2_manhole
    old_col_names = ["display_name", "code", "manhole_indicator",
                     "surface_level", "bottom_level", "drain_level",
                     "calculation_type", "exchange_thickness",
                     "hydraulic_conductivity_in", "hydraulic_conductivity_out"]
    rename_map = {"surface_level": "manhole_surface_level",
                  "bottom_level": "bottom_level",
                  "drain_level": "exchange_level",
                  "calculation_type": "exchange_type",
                  "manhole_indicator": "visualisation"}
    set_items = ',\n'.join(f"""{rename_map.get(col_name, col_name)} = (
        SELECT v2_manhole.{col_name} FROM v2_manhole
        WHERE v2_manhole.connection_node_id = connection_node.id)""" for col_name in old_col_names)
    op.execute(sa.text(f"""
        UPDATE connection_node
        SET {set_items}
        WHERE EXISTS (
            SELECT 1
            FROM v2_manhole
            WHERE v2_manhole.connection_node_id = connection_node.id
        );
    """))


def create_material():
    op.execute(sa.text("""
    CREATE TABLE material (
    id INTEGER PRIMARY KEY NOT NULL,
    description TEXT,
    friction_type INTEGER,
    friction_coefficient REAL);
    """))
    session = Session(bind=op.get_bind())
    nof_settings = session.execute(select(func.count()).select_from(models.ModelSettings)).scalar()
    if nof_settings > 0:
        with open(data_dir.joinpath('0228_materials.csv')) as file:
            reader = csv.DictReader(file)
            session.bulk_save_objects([models.Material(**row) for row in reader])
            session.commit()


def modify_obstacle():
    op.execute(sa.text(f'ALTER TABLE obstacle ADD COLUMN affects_2d BOOLEAN DEFAULT TRUE;'))
    op.execute(sa.text(f'ALTER TABLE obstacle ADD COLUMN affects_1d2d_open_water BOOLEAN DEFAULT TRUE;'))
    op.execute(sa.text(f'ALTER TABLE obstacle ADD COLUMN affects_1d2d_closed BOOLEAN DEFAULT FALSE;'))


def modify_control_target_type():
    for table_name in ['table_control', 'memory_control']:
        op.execute(sa.text(f"""
            UPDATE {table_name}
            SET target_type = REPLACE(target_type, 'v2_pumpstation', 'pump')
            WHERE target_type = 'v2_pumpstation';
        """))
        op.execute(sa.text(f"""
            UPDATE {table_name}
            SET target_type = REPLACE(target_type, 'v2_', '')
            WHERE target_type LIKE 'v2_%';
        """))



def modify_model_settings():
    op.execute(sa.text(f'ALTER TABLE model_settings ADD COLUMN node_open_water_detection INTEGER DEFAULT 1;'))


def check_for_null_geoms():
    tables = ["v2_connection_nodes", "v2_cross_section_location", "v2_culvert", "v2_channel", "v2_windshielding"]
    conn = op.get_bind()
    for table in tables:
        nof_null = conn.execute(sa.text(f"SELECT COUNT(*) FROM {table} WHERE the_geom IS NULL;")).fetchone()[0]
        if nof_null > 0:
            raise Schema228UpgradeException("Cannot migrate because of empty geometries in table {table}")


def fix_material_id():
    # Replace migrated material_id's with correct values
    replace_map = {9 : 2, 10 : 7}
    material_id_tables = ['pipe', 'culvert', 'weir', 'orifice']
    for table in material_id_tables:
        op.execute(sa.text(f"UPDATE {table} SET material_id = CASE material_id "
                           f"{' '.join([f'WHEN {old} THEN {new}' for old, new in replace_map.items()])} "
                           "ELSE material_id END"))



def drop_conflicting():
    new_tables = [new_name for _, new_name in RENAME_TABLES] + ['material', 'pump_map']
    for table_name in new_tables:
        op.execute(f"DROP TABLE IF EXISTS {table_name};")



def upgrade():
    # Empty or non-existing connection node id (start or end) in Orifice, Pipe, Pumpstation or Weir will break
    # migration, so an error is raised in these cases
    check_for_null_geoms()
    # Prevent custom tables in schematisation from breaking migration when they conflict with new table names
    drop_conflicting()
    # Extent cross section definition table (actually stored in temp)
    extend_cross_section_definition_table()
    # Migrate data from cross_section_definition to cross_section_location
    migrate_cross_section_definition_to_location()
    # Prepare object tables for renaming by copying cross section data and setting the_geom
    for table_name in ['v2_culvert', 'v2_weir', 'v2_pipe', 'v2_orifice']:
        migrate_cross_section_definition_to_object(table_name)
        # Set geometry for tables without one
        if table_name != 'v2_culvert':
            set_geom_for_object(table_name)
    set_geom_for_v2_pumpstation()
    for old_table_name, new_table_name in RENAME_TABLES:
        modify_table(old_table_name, new_table_name)
    # Create new tables
    create_pump_map()
    create_material()
    create_connection_node()
    # Modify exsiting tables
    modify_model_settings()
    modify_obstacle()
    modify_control_target_type()
    fix_material_id()
    fix_geometry_columns()
    remove_tables([old for old, _ in RENAME_TABLES]+DELETE_TABLES+[Temp.__tablename__, 'v2_manhole'])


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")



"""Upgrade settings in schema

Revision ID: 0227
Revises:
Create Date: 2024-09-24 15:10

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0227"
down_revision = "0226"
branch_labels = None
depends_on = None

TABLES = ['memory_control', 'table_control']
RENAME_TABLES = [('control_measure_location', 'measure_location'),
                 ('control_measure_map', 'measure_map'), ]


def fix_geometries(downgrade: bool=False):
    op.execute(sa.text("SELECT RecoverGeometryColumn('memory_control', 'geom', 4326, 'POINT', 'XY')"))
    op.execute(sa.text("SELECT RecoverGeometryColumn('table_control', 'geom', 4326, 'POINT', 'XY')"))
    if downgrade:
        op.execute(sa.text("SELECT RecoverGeometryColumn('control_measure_location', 'geom', 4326, 'POINT', 'XY')"))
        op.execute(sa.text("SELECT RecoverGeometryColumn('control_measure_map', 'geom', 4326, 'LINESTRING', 'XY')"))
    else:
        op.execute(sa.text("SELECT RecoverGeometryColumn('measure_location', 'geom', 4326, 'POINT', 'XY')"))
        op.execute(sa.text("SELECT RecoverGeometryColumn('measure_map', 'geom', 4326, 'LINESTRING', 'XY')"))


def upgrade():
    # remove measure variable from memory_control and table_control
    for table_name in TABLES:
        with op.batch_alter_table(table_name) as batch_op:
            batch_op.drop_column('measure_variable')
    # rename column
    with op.batch_alter_table('control_measure_map') as batch_op:
        batch_op.alter_column('control_measure_location_id', new_column_name='measure_location_id')
    # rename tables
    for old_table_name, new_table_name in RENAME_TABLES:
        op.rename_table(old_table_name, new_table_name)
    fix_geometries()


def downgrade():
    # undo remove measure variable from memory_control and table_control
    for table_name in TABLES:
        with op.batch_alter_table(table_name) as batch_op:
            batch_op.add_column(sa.Column("measure_variable", sa.Text, server_default="water_level"))
    # undo rename columns
    with op.batch_alter_table('measure_map') as batch_op:
        batch_op.alter_column('measure_location_id', new_column_name='control_measure_location_id')
    # rename tables
    for old_table_name, new_table_name in RENAME_TABLES:
        op.rename_table(new_table_name, old_table_name)
    fix_geometries(downgrade=True)



"""Removed unused columns.

Revision ID: 0203
Revises:
Create Date: 2021-09-29 13:50:19.544275

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0203"
down_revision = "0202"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("v2_aggregation_settings") as batch_op:
        batch_op.drop_column("aggregation_in_space")


def downgrade():
    with op.batch_alter_table("v2_aggregation_settings") as batch_op:
        batch_op.add_column(
            sa.Column("aggregation_in_space", sa.Boolean(), nullable=False),
        )



"""Upgrade settings in schema

Revision ID: 0223
Revises:
Create Date: 2024-05-27 10:35

"""
import json
import warnings
from pathlib import Path
from typing import Dict, List, Tuple

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Boolean, Column, Float, Integer, String, Text
from sqlalchemy.event import listen
from sqlalchemy.orm import declarative_base

from threedi_schema.application.threedi_database import load_spatialite
from threedi_schema.domain.custom_types import Geometry

# revision identifiers, used by Alembic.
revision = "0223"
down_revision = "0222"
branch_labels = None
depends_on = None

Base = declarative_base()

data_dir = Path(__file__).parent / "data"

# (source table, destination table)
RENAME_TABLES = [
    ("v2_surface_parameters", "surface_parameters"),
]

ADD_COLUMNS = [
    ("surface_parameters", Column("description", Text)),
    ("surface_parameters", Column("tags", Text)),
]

ADD_TABLES = {
    "surface": [
        Column("area", Float),
        Column("surface_parameters_id", Integer, default=1),
        Column("tags", Text),
        Column("code", String(100)),
        Column("display_name", String(255)),
    ],
    "dry_weather_flow": [
        Column("multiplier", Float),
        Column("dry_weather_flow_distribution_id", Integer, default=1),
        Column("daily_total", Float),
        Column("interpolate", Boolean, default=False),
        Column("tags", Text),
        Column("code", String(100)),
        Column("display_name", String(255)),
    ],
    "surface_map": [
        Column("connection_node_id", Integer),
        Column("surface_id", Integer),
        Column("percentage", Float),
        Column("tags", Text),
        Column("code", String(100)),
        Column("display_name", String(255)),
    ],
    "dry_weather_flow_map": [
        Column("connection_node_id", Integer),
        Column("dry_weather_flow_id", Integer),
        Column("percentage", Float),
        Column("tags", Text),
        Column("code", String(100)),
        Column("display_name", String(255)),
    ],
    "dry_weather_flow_distribution": [
        Column("description", Text),
        Column("tags", Text),
        Column("distribution", Text)
    ],
    "tags": [
        Column("description", Text)
    ]
}

# Geom columns need to be added using geoalchemy, so therefore that's a seperate task
NEW_GEOM_COLUMNS = {
    ("surface", Column("geom", Geometry("POLYGON"), nullable=False)),
    ("dry_weather_flow", Column("geom", Geometry("POLYGON"), nullable=False)),
    ("surface_map", Column("geom", Geometry("LINESTRING"), nullable=False)),
    ("dry_weather_flow_map", Column("geom", Geometry("LINESTRING"), nullable=False))
}

REMOVE_TABLES = [
    "v2_impervious_surface",
    "v2_impervious_surface_map",
    "v2_surface",
    "v2_surface_map"
]


class NoMappingWarning(UserWarning):
    pass


def rename_tables(table_sets: List[Tuple[str, str]]):
    # no checks for existence are done, this will fail if a source table doesn't exist
    for src_name, dst_name in table_sets:
        op.rename_table(src_name, dst_name)


def create_new_tables(new_tables: Dict[str, sa.Column]):
    # no checks for existence are done, this will fail if any table already exists
    for table_name, columns in new_tables.items():
        op.create_table(table_name, sa.Column("id", sa.Integer(), primary_key=True),
                        *columns)


def add_columns_to_tables(table_columns: List[Tuple[str, Column]]):
    # no checks for existence are done, this will fail if any column already exists
    for dst_table, col in table_columns:
        if isinstance(col.type, Geometry):
            add_geometry_column(dst_table, col)
        else:
            with op.batch_alter_table(dst_table) as batch_op:
                batch_op.add_column(col)


def add_geometry_column(table: str, geocol: Column):
    # Adding geometry columns via alembic doesn't work
    # https://postgis.net/docs/AddGeometryColumn.html
    geotype = geocol.type
    query = (
        f"SELECT AddGeometryColumn('{table}', '{geocol.name}', {geotype.srid}, '{geotype.geometry_type}', 'XY', 0);")
    op.execute(sa.text(query))


def remove_tables(tables: List[str]):
    for table in tables:
        op.drop_table(table)


def copy_values_to_new_table(src_table: str, src_columns: List[str], dst_table: str, dst_columns: List[str]):
    query = f'INSERT INTO {dst_table} ({", ".join(dst_columns)}) SELECT {", ".join(src_columns)} FROM {src_table}'
    op.execute(sa.text(query))


def copy_v2_data_to_surface(src_table: str):
    src_columns = ["id", "code", "display_name", "sur_geom", "area"]
    dst_columns = ["id", "code", "display_name", "geom", "area"]
    if src_table == "v2_surface":
        src_columns += ["surface_parameters_id"]
        dst_columns += ["surface_parameters_id"]
    copy_values_to_new_table(src_table, src_columns, "surface", dst_columns)
    op.execute(sa.text("DELETE FROM surface WHERE area = 0 OR area IS NULL;"))


def copy_v2_data_to_dry_weather_flow(src_table: str):
    src_columns = ["id", "code", "display_name", "dwf_geom", "nr_of_inhabitants", "dry_weather_flow"]
    dst_columns = ["id", "code", "display_name", "geom", "multiplier", "daily_total"]
    copy_values_to_new_table(src_table, src_columns, "dry_weather_flow", dst_columns)
    op.execute(sa.text("""
    UPDATE dry_weather_flow
    SET dry_weather_flow_distribution_id = 1,
        interpolate = 0;
        """))
    op.execute(sa.text("DELETE FROM dry_weather_flow "
                       "WHERE multiplier = 0 OR daily_total = 0 OR multiplier IS NULL OR daily_total IS NULL;"))


def remove_orphans_from_map(basename: str):
    query = f"DELETE FROM {basename}_map WHERE {basename}_id NOT IN (SELECT id FROM {basename});"
    op.execute(sa.text(query))


def copy_v2_data_to_dry_weather_flow_map(src_table: str):
    src_columns = ["connection_node_id", "percentage", src_table.strip('v2_').replace('_map', '_id')]
    dst_columns = ["connection_node_id", "percentage", "dry_weather_flow_id"]
    copy_values_to_new_table(src_table, src_columns, "dry_weather_flow_map", dst_columns)


def copy_v2_data_to_surface_map(src_table: str):
    src_columns = ["connection_node_id", "percentage", src_table.strip('v2_').replace('_map', '_id')]
    dst_columns = ["connection_node_id", "percentage", "surface_id"]
    copy_values_to_new_table(src_table, src_columns, "surface_map", dst_columns)


def set_map_geometries(basename):
    # Set geom as a line between point on surface/dry_weather_flow and connection node
    query = f"""
        UPDATE {basename}_map AS map
        SET geom = (
            SELECT MakeLine(PointOnSurface(obj.geom), vcn.the_geom)
            FROM {basename} obj
            JOIN v2_connection_nodes vcn ON map.connection_node_id = vcn.id
            WHERE obj.id = map.{basename}_id
        );        
    """
    op.execute(sa.text(query))


def add_map_geometries(src_table: str):
    # Add geometries to a map table that connects the connection node and the surface / dry_weather_flow
    srid = get_global_srid()
    query = f"""
        UPDATE {src_table}_map
        SET geom = (
            SELECT CASE
                WHEN ST_Equals(c.the_geom, PointOnSurface(s.geom)) THEN
                    -- Transform to EPSG:4326 for the projection, then back to the original SRID
                    MakeLine(
                        c.the_geom,
                        PointOnSurface(ST_Transform(
                            ST_Translate(
                                ST_Transform(s.geom, {srid}),
                                0, 1, 0
                            ),
                            4326
                        ))                       
                    )
                ELSE
                    MakeLine(c.the_geom, PointOnSurface(s.geom))
                END                
            FROM v2_connection_nodes c, {src_table} s
            WHERE c.id = {src_table}_map.connection_node_id 
            AND s.id = {src_table}_map.{src_table}_id
        )
        WHERE EXISTS (
            SELECT 1
            FROM v2_connection_nodes c, {src_table} s
            WHERE c.id = {src_table}_map.connection_node_id 
            AND s.id = {src_table}_map.{src_table}_id
        );
    """
    op.execute(sa.text(query))



def get_global_srid():
    conn = op.get_bind()
    use_0d_inflow = conn.execute(sa.text("SELECT use_0d_inflow FROM simulation_template_settings LIMIT 1")).fetchone()
    if use_0d_inflow is not None:
        srid = conn.execute(sa.text("SELECT epsg_code FROM model_settings LIMIT 1")).fetchone()
        if (srid is not None) and (srid[0] is not None):
            return srid[0]
    return 28992


def get_area_str(geom_str) -> str:
    # Get SQLite statement to compute area for a given geometry
    return f'ST_Area(ST_Transform({geom_str},{get_global_srid()}))'


def copy_polygons(src_table: str, tmp_geom: str):
    # Copy existing polygons in src_table to a new column (tmp_geom):
    # - directly copy polygons
    # - copy the first item of all multipolygons
    # - add new rows for each extra polygon inside a multipolygon
    conn = op.get_bind()
    # Copy polygons directly
    op.execute(sa.text(f"UPDATE {src_table} SET {tmp_geom} = the_geom WHERE GeometryType(the_geom) = 'POLYGON';"))
    # Copy first polygon of each multipolygon and correct the area
    op.execute(sa.text(f"""
    UPDATE {src_table} 
    SET {tmp_geom} = ST_GeometryN(the_geom,1), area = {get_area_str('ST_GeometryN(the_geom,1)')}
    WHERE GeometryType(the_geom) = 'MULTIPOLYGON' 
    AND GeometryType(ST_GeometryN(the_geom,1))  = 'POLYGON';
    """))
    # Copy the remaining polygons for multipolygons with more than one polygon
    # select column names that we will copy directly
    col_names = [col_info[1] for col_info in
                 conn.execute(sa.text(f"PRAGMA table_info({src_table})")).fetchall()]
    col_str = ', '.join(list(set(col_names) - {'id', 'the_geom', 'tmp_geom', 'area'}))
    conn = op.get_bind()
    rows = conn.execute(sa.text(f"""
        SELECT id, {col_str}, NumGeometries(the_geom) FROM {src_table}
        WHERE GeometryType(the_geom) = 'MULTIPOLYGON'
        AND GeometryType(ST_GeometryN(the_geom,1))  = 'POLYGON'
        AND NumGeometries(the_geom) > 1;""")).fetchall()
    id_next = conn.execute(sa.text(f"SELECT MAX(id) FROM {src_table}")).fetchall()[0][0]
    surf_id = f"{src_table.strip('v2_')}_id"
    for row in rows:
        id = row[0]
        nof_polygons = row[-1]
        # Retrieve data from map table
        conn_node_id, percentage = conn.execute(sa.text(f"""
            SELECT connection_node_id, percentage FROM {src_table}_map
            WHERE {surf_id} = {id}""")).fetchall()[0]
        for i in range(2, nof_polygons + 1):
            id_next += 1
            # Copy polygon to new row
            op.execute(sa.text(f"""
                INSERT INTO {src_table} (id, the_geom, {tmp_geom}, area, {col_str})
                SELECT {id_next}, the_geom, ST_GeometryN(the_geom, {i}),
                {get_area_str(f'ST_GeometryN(the_geom,{i})')}, {col_str}
                FROM {src_table} WHERE id = {id} LIMIT 1
            """))
            # Add new row to the map
            op.execute(sa.text(f"""
                INSERT INTO {src_table}_map ({surf_id}, connection_node_id, percentage)
                VALUES ({id_next}, {conn_node_id}, {percentage})
            """))


def create_buffer_polygons(src_table: str, tmp_geom: str):
    # create circular polygon of area 1 around the connection node
    surf_id = f"{src_table.strip('v2_')}_id"
    op.execute(sa.text(f"""
    UPDATE {src_table} 
    SET {tmp_geom} = (
        SELECT ST_Buffer(v2_connection_nodes.the_geom, 1)
        FROM v2_connection_nodes 
        JOIN {src_table}_map 
        ON v2_connection_nodes.id = {src_table}_map.connection_node_id
        WHERE {src_table}.id = {src_table}_map.{surf_id}
    )
    WHERE {tmp_geom} IS NULL
    AND id IN (
        SELECT {src_table}_map.{surf_id}
        FROM v2_connection_nodes 
        JOIN {src_table}_map 
        ON v2_connection_nodes.id = {src_table}_map.connection_node_id
    );
    """))


def create_square_polygons(src_table: str, tmp_geom: str):
    # create square polygon with area area around the connection node
    side_expr = f'sqrt({src_table}.area)'
    surf_id = f"{src_table.strip('v2_')}_id"
    # When no geometry is defined, a square with area matching the area column
    # with the center at the connection node is added
    srid = get_global_srid()
    query_str = f"""
        WITH center AS (
          SELECT {src_table}.id AS item_id,
                    ST_Centroid(ST_Collect(
                        ST_Transform(v2_connection_nodes.the_geom, {srid}))) AS geom
          FROM {src_table}_map
          JOIN v2_connection_nodes ON {src_table}_map.connection_node_id = v2_connection_nodes.id
          JOIN {src_table} ON {src_table}_map.{surf_id} = {src_table}.id    
          WHERE {src_table}_map.{surf_id} = {src_table}.id
          GROUP BY {src_table}.id
        ),
        side_length AS (
          SELECT {side_expr} AS side
        )
        UPDATE {src_table}
        SET {tmp_geom} = (
            SELECT ST_Transform(
                    SetSRID(
                        ST_GeomFromText('POLYGON((' ||
                            (ST_X(center.geom) - side_length.side / 2) || ' ' || (ST_Y(center.geom) - side_length.side / 2) || ',' ||
                            (ST_X(center.geom) + side_length.side / 2) || ' ' || (ST_Y(center.geom) - side_length.side / 2) || ',' ||
                            (ST_X(center.geom) + side_length.side / 2) || ' ' || (ST_Y(center.geom) + side_length.side / 2) || ',' ||
                            (ST_X(center.geom) - side_length.side / 2) || ' ' || (ST_Y(center.geom) + side_length.side / 2) || ',' ||
                            (ST_X(center.geom) - side_length.side / 2) || ' ' || (ST_Y(center.geom) - side_length.side / 2) ||
                            '))'),
                        {srid}),
                4326
            ) AS transformed_geom
            FROM center, side_length
            WHERE center.item_id = {src_table}.id
        )
        WHERE {tmp_geom} IS NULL;
        """
    op.execute(sa.text(query_str))

def fix_src_geometry(src_table: str, tmp_geom: str, create_polygons):
    conn = op.get_bind()
    # create columns to store the derived geometries to
    op.execute(sa.text(f"SELECT AddGeometryColumn('{src_table}', '{tmp_geom}', 4326, 'POLYGON', 'XY', 0);"))
    # Copy existing polygons
    copy_polygons(src_table, tmp_geom)
    # Check if any existing geometries where not copied
    not_copied = conn.execute(sa.text(f'SELECT id FROM {src_table} '
                                      f'WHERE {tmp_geom} IS NULL '
                                      f'AND the_geom IS NOT NULL')).fetchall()
    if len(not_copied) > 0:
        raise BaseException(f'Found {len(not_copied)} geometries in {src_table} that could not'
                            f'be converted to a POLYGON geometry: {not_copied}')
    # Create polygons for rows where no geometry was defined
    create_polygons(src_table, tmp_geom)


def remove_invalid_rows(src_table:str):
    # Remove rows with insufficient data
    op.execute(sa.text(f"DELETE FROM {src_table} WHERE area = 0 "
                       "AND (nr_of_inhabitants = 0 OR dry_weather_flow = 0);"))

    # Remove rows without mapping
    conn = op.get_bind()
    where_clause = (f"WHERE id NOT IN (SELECT {src_table.strip('v2_')}_id FROM {src_table}_map) "
                    f"AND the_geom IS NULL")
    no_map_id = conn.execute(sa.text(f"SELECT id FROM {src_table} {where_clause};")).fetchall()
    if len(no_map_id) > 0:
        op.execute(sa.text(f"DELETE FROM {src_table} {where_clause};"))
        msg = (f"Could not migrate the following rows from {src_table} because "
               f"they are not mapped to a connection node in {src_table}_map: {no_map_id}")
        warnings.warn(msg, NoMappingWarning)

def populate_surface_and_dry_weather_flow():
    conn = op.get_bind()
    use_0d_inflow = conn.execute(sa.text("SELECT use_0d_inflow FROM simulation_template_settings LIMIT 1")).fetchone()
    if (use_0d_inflow is None) or (len(use_0d_inflow) == 0) or (use_0d_inflow[0] not in [1, 2]):
        return
    use_0d_inflow = use_0d_inflow[0]
    # Use use_0d_inflow setting to determine wether to copy any data and if so from what table
    src_table = "v2_impervious_surface" if use_0d_inflow == 1 else "v2_surface"
    remove_invalid_rows(src_table)

    # Create geometries for non-specified ones
    # Add geometries for surfaces and dwf by adding extra columns
    # This has to be done in advance because NULL geometries cannot be copied
    # And this had to be done seperately because the geometries for surfaces and
    # DWF are not by definition the same
    fix_src_geometry(src_table, 'sur_geom', create_square_polygons)
    fix_src_geometry(src_table, 'dwf_geom', create_buffer_polygons)
    # Copy data to new tables
    copy_v2_data_to_surface(src_table)
    copy_v2_data_to_dry_weather_flow(src_table)
    copy_v2_data_to_surface_map(f"{src_table}_map")
    copy_v2_data_to_dry_weather_flow_map(f"{src_table}_map")

    # Remove rows in maps that refer to non-existing objects
    remove_orphans_from_map(basename="surface")
    remove_orphans_from_map(basename="dry_weather_flow")

    # Create geometries in new maps
    add_map_geometries("surface")
    add_map_geometries("dry_weather_flow")

    # Set surface parameter id
    if use_0d_inflow == 1:
        set_surface_parameters_id()
    # Populate tables with default values
    populate_dry_weather_flow_distribution()
    populate_surface_parameters()


def set_surface_parameters_id():
    # Make sure not to call this on an empty database
    with open(data_dir.joinpath('0223_surface_parameters_map.json'), 'r') as f:
        parameter_map = json.load(f)
    conn = op.get_bind()
    res = conn.execute(
        sa.text("SELECT surface_class, surface_inclination FROM v2_impervious_surface")).fetchone()
    if res is None:
        return
    surface_class, surface_inclination = res
    parameter_id = parameter_map[f'{surface_class} - {surface_inclination}']
    op.execute(f'UPDATE surface SET surface_parameters_id = {parameter_id}')


def populate_surface_parameters():
    # Make sure not to call this on an empty database
    with open(data_dir.joinpath('0223_surface_parameters_contents.json'), 'r') as f:
        data_to_insert = json.load(f)
    keys_str = "(" + ",".join(data_to_insert[0].keys()) + ")"
    for row in data_to_insert:
        val_str = "(" + ",".join([repr(item) for item in row.values()]) + ")"
        sql_query = f"INSERT INTO surface_parameters {keys_str} VALUES {val_str}"
        op.execute(sa.text(sql_query))


def populate_dry_weather_flow_distribution():
    with open(data_dir.joinpath('0223_dry_weather_flow_distribution.csv'), 'r') as f:
        distr = f.read().strip()
    description = "Kennisbank Stichting Rioned - https://www.riool.net/huishoudelijk-afvalwater"
    sql_query = f"INSERT INTO dry_weather_flow_distribution (description, distribution) VALUES ('{description}', '{distr}')"
    op.execute(sa.text(sql_query))


def fix_geometry_columns():
    GEO_COL_INFO = [
        ('dry_weather_flow', 'geom', 'POLYGON'),
        ('dry_weather_flow_map', 'geom', 'LINESTRING'),
        ('surface', 'geom', 'POLYGON'),
        ('surface_map', 'geom', 'LINESTRING'),
    ]
    for table, column, geotype in GEO_COL_INFO:
        with op.batch_alter_table(table) as batch_op:
            batch_op.alter_column(column_name=column, nullable=False)
        migration_query = f"SELECT RecoverGeometryColumn('{table}', '{column}', {4326}, '{geotype}', 'XY')"
        op.execute(sa.text(migration_query))


def drop_conflicting():
    new_tables = list(ADD_TABLES.keys()) + [new_name for _, new_name in RENAME_TABLES]
    for table_name in new_tables:
        op.execute(f"DROP TABLE IF EXISTS {table_name};")


def upgrade():
    connection = op.get_bind()
    listen(connection.engine, "connect", load_spatialite)
    # Remove existing tables (outside of the specs) that conflict with new table names
    drop_conflicting()
    # create new tables and rename existing tables
    create_new_tables(ADD_TABLES)
    rename_tables(RENAME_TABLES)
    # add new columns to existing tables
    add_columns_to_tables(ADD_COLUMNS)
    add_columns_to_tables(NEW_GEOM_COLUMNS)
    # migrate values from old tables to new tables
    populate_surface_and_dry_weather_flow()
    # recover geometry columns
    fix_geometry_columns()
    # remove old tables
    remove_tables(REMOVE_TABLES)


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")
